import json
import os
import pathlib
import time
from typing import Optional, List, Set, Dict, Union, TYPE_CHECKING

import git
from labml import logger, monit
from labml.internal.configs.base import Configs
from labml.internal.configs.processor import ConfigProcessor, FileConfigsSaver
from labml.internal.experiment.experiment_run import Run, struct_time_to_time, struct_time_to_date
from labml.internal.experiment.watcher import ExperimentWatcher
from labml.internal.lab import lab_singleton
from labml.internal.monitor import monitor_singleton as monitor
from labml.internal.tracker import tracker_singleton as tracker
from labml.internal.util import is_ipynb, is_colab, is_kaggle
from labml.logger import Text
from labml.utils import get_caller_file
from labml.utils.notice import labml_notice

if TYPE_CHECKING:
    from labml.internal.api.experiment import ApiExperiment


class ModelSaver:
    def save(self, checkpoint_path: pathlib.Path) -> any:
        raise NotImplementedError()

    def load(self, checkpoint_path: pathlib.Path, info: any):
        raise NotImplementedError()


class CheckpointSaver:
    model_savers: Dict[str, ModelSaver]

    def __init__(self, path: pathlib.PurePath):
        self.path = path
        self.model_savers = {}
        self.__no_savers_warned = False

    def add_savers(self, models: Dict[str, ModelSaver]):
        """
        ## Set variable for saving and loading
        """
        if experiment_singleton().is_started:
            raise RuntimeError('Cannot register models with the experiment after experiment has started.'
                               'Register models before calling experiment.start')

        self.model_savers.update(models)

    def save(self, global_step):
        """
        ## Save model as a set of numpy arrays
        """

        if not self.model_savers:
            if not self.__no_savers_warned:
                labml_notice(["No models were registered for saving\n",
                              "You can register models with ",
                              ('experiment.add_pytorch_models', Text.value)])
                self.__no_savers_warned = True
            return

        checkpoints_path = pathlib.Path(self.path)
        if not checkpoints_path.exists():
            checkpoints_path.mkdir()

        checkpoint_path = checkpoints_path / str(global_step)
        assert not checkpoint_path.exists()

        checkpoint_path.mkdir()

        info = {}
        for name, saver in self.model_savers.items():
            info[name] = saver.save(checkpoint_path)

        # Save header
        with open(str(checkpoint_path / "info.json"), "w") as f:
            f.write(json.dumps(info))

    def load(self, checkpoint_path: pathlib.Path, models: List[str] = None):
        """
        ## Load model as a set of numpy arrays
        """

        if not self.model_savers:
            if not self.__no_savers_warned:
                labml_notice(["No models were registered for loading or saving\n",
                              "You can register models with ",
                              ('experiment.add_pytorch_models', Text.value)])
                self.__no_savers_warned = True
            return

        if not models:
            models = list(self.model_savers.keys())

        with open(str(checkpoint_path / "info.json"), "r") as f:
            info = json.loads(f.readline())

        to_load = []
        not_loaded = []
        missing = []
        for name in models:
            if name not in info:
                missing.append(name)
            else:
                to_load.append(name)
        for name in info:
            if name not in models:
                not_loaded.append(name)

        # Load each model
        for name in to_load:
            saver = self.model_savers[name]
            saver.load(checkpoint_path, info[name])

        if missing:
            labml_notice([(f'{missing} ', Text.highlight),
                          ('model(s) could not be found.\n'),
                          (f'{to_load} ', Text.none),
                          ('models were loaded.', Text.none)
                          ], is_danger=True)
        if not_loaded:
            labml_notice([(f'{not_loaded} ', Text.none),
                          ('models were not loaded.\n', Text.none),
                          'Models to be loaded should be specified with: ',
                          ('experiment.add_pytorch_models', Text.value)])


class Experiment:
    r"""
    Each experiment has different configurations or algorithms.
    An experiment can have multiple runs.

    Keyword Arguments:
        name (str, optional): name of the experiment
        python_file (str, optional): path of the Python file that
            created the experiment
        comment (str, optional): a short description of the experiment
        writers (Set[str], optional): list of writers to write stat to
        ignore_callers: (Set[str], optional): list of files to ignore when
            automatically determining ``python_file``
        tags (Set[str], optional): Set of tags for experiment
    """
    web_api: Optional['ApiExperiment']
    is_started: bool
    run: Run
    configs_processor: Optional[ConfigProcessor]

    # whether not to start the experiment if there are uncommitted changes.
    check_repo_dirty: bool
    checkpoint_saver: CheckpointSaver

    distributed_rank: int
    distributed_world_size: int

    def __init__(self, *,
                 uuid: str,
                 name: Optional[str],
                 python_file: Optional[str],
                 comment: Optional[str],
                 writers: Set[str],
                 ignore_callers: Set[str],
                 tags: Optional[Set[str]],
                 is_evaluate: bool):
        if python_file is None:
            python_file = get_caller_file(ignore_callers)

        if python_file.startswith('<ipython'):
            assert is_ipynb()
            if name is None:
                raise ValueError("You must specify python_file or experiment name"
                                 " when creating an experiment from a python notebook.")

            lab_singleton().set_path(os.getcwd())
            python_file = 'notebook.ipynb'
        else:
            lab_singleton().set_path(python_file)

            if name is None:
                file_path = pathlib.PurePath(python_file)
                name = file_path.stem

        if comment is None:
            comment = ''
        if global_params_singleton().comment is not None:
            comment = global_params_singleton().comment

        self.experiment_path = lab_singleton().experiments / name

        self.check_repo_dirty = lab_singleton().check_repo_dirty

        self.configs_processor = None

        if tags is None:
            tags = set(name.split('_'))

        self.run = Run.create(
            uuid=uuid,
            experiment_path=self.experiment_path,
            python_file=python_file,
            trial_time=time.localtime(),
            name=name,
            comment=comment,
            tags=list(tags))

        try:
            repo = git.Repo(lab_singleton().path)

            self.run.commit = repo.head.commit.hexsha
            self.run.commit_message = repo.head.commit.message.strip()
            self.run.is_dirty = repo.is_dirty()
            self.run.diff = repo.git.diff()
        except git.InvalidGitRepositoryError:
            if not is_colab() and not is_kaggle():
                labml_notice(["Not a valid git repository: ",
                              (lab_singleton().path, Text.value)])
            self.run.commit = 'unknown'
            self.run.commit_message = ''
            self.run.is_dirty = True
            self.run.diff = ''

        self.checkpoint_saver = CheckpointSaver(self.run.checkpoint_path)
        self.is_evaluate = is_evaluate
        self.web_api = None
        self.writers = writers
        self.is_started = False
        self.distributed_rank = 0
        self.distributed_world_size = -1

    def __print_info(self):
        """
        🖨 Print the experiment info and check git repo status
        """

        logger.log()
        logger.log([
            (self.run.name, Text.title),
            ': ',
            (str(self.run.uuid), Text.meta)
        ])

        if self.run.comment != '':
            logger.log(['\t', (self.run.comment, Text.highlight)])

        commit_message = self.run.commit_message.strip().replace('\n', '¶ ').replace('\r', '')
        logger.log([
            "\t"
            "[dirty]" if self.run.is_dirty else "[clean]",
            ": ",
            (f"\"{commit_message}\"", Text.highlight)
        ])

        if self.run.load_run is not None:
            logger.log([
                "\t"
                "loaded from",
                ": ",
                (f"{self.run.load_run}", Text.meta2),
            ])

    def _load_checkpoint(self, checkpoint_path: pathlib.Path):
        self.checkpoint_saver.load(checkpoint_path)

    def save_checkpoint(self):
        if self.is_evaluate:
            return
        if self.distributed_rank != 0:
            return

        self.checkpoint_saver.save(tracker().global_step)

    def calc_configs(self,
                     configs: Union[Configs, Dict[str, any]],
                     configs_override: Optional[Dict[str, any]]):
        if configs_override is None:
            configs_override = {}
        if global_params_singleton().configs is not None:
            configs_override.update(global_params_singleton().configs)

        self.configs_processor = ConfigProcessor(configs, configs_override)

        if self.distributed_rank == 0:
            logger.log()

    def __start_from_checkpoint(self, run_uuid: str, checkpoint: Optional[int]):
        checkpoint_path, global_step = experiment_run.get_run_checkpoint(
            run_uuid,
            checkpoint)

        if global_step is None:
            return 0
        else:
            with monit.section("Loading checkpoint"):
                self._load_checkpoint(checkpoint_path)
            self.run.load_run = run_uuid

        return global_step

    def load_models(self, *,
                    models: List[str],
                    run_uuid: Optional[str] = None,
                    checkpoint: Optional[int] = None):
        if checkpoint is None:
            checkpoint = -1
        checkpoint_path, global_step = experiment_run.get_run_checkpoint(run_uuid, checkpoint)

        if global_step is None:
            labml_notice(['Could not find saved checkpoint'], is_danger=True)
            return

        with monit.section("Loading checkpoint"):
            self.checkpoint_saver.load(checkpoint_path, models)

    def _save_pid(self):
        if not self.run.pids_path.exists():
            self.run.pids_path.mkdir()

        pid_path = self.run.pids_path / f'{self.distributed_rank}.pid'
        assert not pid_path.exists()

        with open(str(pid_path), 'w') as f:
            f.write(f'{os.getpid()}')

    def distributed(self, rank: int, world_size: int):
        self.distributed_rank = rank
        self.distributed_world_size = world_size

        if self.distributed_rank != 0:
            monitor().silent()

        # to make sure we have the path to save pid
        self.run.make_path()

    def _start_tracker(self):
        tracker().reset_writers()

        if self.is_evaluate:
            return

        if self.distributed_rank != 0:
            return

        if 'screen' in self.writers:
            from labml.internal.tracker.writers import screen
            tracker().add_writer(screen.ScreenWriter())

        if 'sqlite' in self.writers:
            from labml.internal.tracker.writers import sqlite
            tracker().add_writer(sqlite.Writer(self.run.sqlite_path, self.run.artifacts_folder))

        if 'tensorboard' in self.writers:
            from labml.internal.tracker.writers import tensorboard
            tracker().add_writer(tensorboard.Writer(self.run.tensorboard_log_path))

        if 'file' in self.writers:
            from labml.internal.tracker.writers import file
            tracker().add_writer(file.Writer(self.run.log_file))

        if 'web_api' in self.writers:
            web_api_conf = lab_singleton().web_api
            if web_api_conf is not None:
                from labml.internal.tracker.writers import web_api
                from labml.internal.api import ApiCaller
                from labml.internal.api.experiment import ApiExperiment
                api_caller = ApiCaller(web_api_conf.url,
                                       {'run_uuid': self.run.uuid},
                                       15)
                self.web_api = ApiExperiment(api_caller,
                                             frequency=web_api_conf.frequency,
                                             open_browser=web_api_conf.open_browser)
                tracker().add_writer(web_api.Writer(api_caller,
                                                    frequency=web_api_conf.frequency))
        else:
            self.web_api = None

    def start(self, *,
              run_uuid: Optional[str] = None,
              checkpoint: Optional[int] = None):
        if run_uuid is not None:
            if checkpoint is None:
                checkpoint = -1
            global_step = self.__start_from_checkpoint(run_uuid, checkpoint)
        else:
            global_step = 0

        self.run.start_step = global_step

        self._start_tracker()
        tracker().set_start_global_step(global_step)

        if self.distributed_rank == 0:
            self.__print_info()
            if self.check_repo_dirty and self.run.is_dirty:
                logger.log([("[FAIL]", Text.danger),
                            " Cannot trial an experiment with uncommitted changes."])
                exit(1)

        if not self.is_evaluate:
            if self.distributed_rank == 0:
                self.run.save_info()
            self._save_pid()

            if self.distributed_rank == 0:
                if self.configs_processor is not None:
                    self.configs_processor.add_saver(FileConfigsSaver(self.run.configs_path))

                if self.web_api is not None:
                    self.web_api.set_info(run_uuid=self.run.uuid,
                                          name=self.run.name,
                                          comment=self.run.comment)
                    self.web_api.start()
                    if self.configs_processor is not None:
                        self.configs_processor.add_saver(self.web_api.get_configs_saver())

                tracker().save_indicators(self.run.indicators_path)

                # PERF: Writing to tensorboard takes about 4 seconds
                # Also wont work when configs are updated live
                # if self.configs_processor:
                #     tracker().write_h_parameters(self.configs_processor.get_hyperparams())

        self.is_started = True
        return ExperimentWatcher(self)

    def finish(self, status: str, details: any = None):
        if not self.is_evaluate:
            with open(str(self.run.run_log_path), 'a') as f:
                end_time = time.time()
                data = json.dumps({'status': status,
                                   'rank': self.distributed_rank,
                                   'details': details,
                                   'time': end_time}, indent=None)
                f.write(data + '\n')

        tracker().finish_loop()

        if self.web_api is not None:
            self.web_api.status(self.distributed_rank, status, details, end_time)


class GlobalParams:
    def __init__(self):
        self.configs = None
        self.comment = None


_global_params: Optional[GlobalParams] = None
_internal: Optional[Experiment] = None


def global_params_singleton() -> GlobalParams:
    global _global_params

    if _global_params is None:
        _global_params = GlobalParams()

    return _global_params


def has_experiment() -> bool:
    global _internal

    return _internal is not None


def experiment_singleton() -> Experiment:
    global _internal

    assert _internal is not None

    return _internal


def create_experiment(*,
                      uuid: str,
                      name: Optional[str],
                      python_file: Optional[str],
                      comment: Optional[str],
                      writers: Set[str],
                      ignore_callers: Set[str],
                      tags: Optional[Set[str]],
                      is_evaluate: bool):
    global _internal

    _internal = Experiment(uuid=uuid,
                           name=name,
                           python_file=python_file,
                           comment=comment,
                           writers=writers,
                           ignore_callers=ignore_callers,
                           tags=tags,
                           is_evaluate=is_evaluate)



from typing import Iterable
import numpy as np


def is_pattern_match(key: str, pattern: str):
    dp = np.zeros((len(key) + 1, len(pattern) + 1), dtype=np.bool)
    dp[0, 0] = True

    for i in range(len(key)):
        for j in range(len(pattern)):
            if key[i] == pattern[j]:
                dp[i + 1][j + 1] = dp[i][j]
            elif pattern[j] == '?':
                dp[i + 1][j + 1] = dp[i][j]
            elif pattern[j] == '*':
                dp[i + 1][j + 1] = dp[i][j] or dp[i][j + 1] or dp[i + 1][j]

    return bool(dp[len(key), len(pattern)])


def find_best_pattern(key: str, patterns: Iterable[str]):
    max_score = -1
    best = None
    for p in patterns:
        if is_pattern_match(key, p):
            s = 0
            for c in p:
                if c not in {'*', '?'}:
                    s += 1

            if s > max_score:
                max_score = s
                best = p

    return best, max_score / len(key)



from pathlib import PurePath, Path
from typing import List, Callable, Dict, Optional

import torch
from torch.utils.data import IterableDataset

from labml import monit


class TextDataset:
    itos: List[str]
    stoi: Dict[str, int]
    n_tokens: int
    train: str
    valid: str
    standard_tokens: List[str] = []

    @staticmethod
    def load(path: PurePath):
        with open(str(path), 'r') as f:
            return f.read()

    def __init__(self, path: PurePath, tokenizer: Callable, train: str, valid: str, test: str):
        self.test = test
        self.valid = valid
        self.train = train
        self.tokenizer = tokenizer
        self.path = path

        self.n_tokens = len(self.standard_tokens)
        self.stoi = {t: i for i, t in enumerate(self.standard_tokens)}

        with monit.section("Tokenize"):
            tokens = self.tokenizer(self.train) + self.tokenizer(self.valid)
            tokens = sorted(list(set(tokens)))

        for t in monit.iterate("Build vocabulary", tokens):
            self.stoi[t] = self.n_tokens
            self.n_tokens += 1

        self.itos = [''] * self.n_tokens
        for t, n in self.stoi.items():
            self.itos[n] = t

    def text_to_i(self, text: str) -> torch.Tensor:
        tokens = self.tokenizer(text)
        return torch.tensor([self.stoi[s] for s in tokens], dtype=torch.long)

    def __repr__(self):
        return f'{len(self.train) / 1_000_000 :,.2f}M, {len(self.valid) / 1_000_000 :,.2f}M - {str(self.path)}'


class SequentialDataLoader(IterableDataset):
    def __init__(self, *, text: str, dataset: TextDataset,
                 batch_size: int, seq_len: int):
        self.seq_len = seq_len
        data = dataset.text_to_i(text)
        n_batch = data.shape[0] // batch_size
        data = data.narrow(0, 0, n_batch * batch_size)
        data = data.view(batch_size, -1).t().contiguous()
        self.data = data

    def __len__(self):
        return self.data.shape[0] // self.seq_len

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx >= self.data.shape[0] - 1:
            raise StopIteration()

        seq_len = min(self.seq_len, self.data.shape[0] - 1 - self.idx)
        i = self.idx + seq_len
        data = self.data[self.idx: i]
        target = self.data[self.idx + 1: i + 1]
        self.idx = i
        return data, target

    def __getitem__(self, idx):
        seq_len = min(self.seq_len, self.data.shape[0] - 1 - idx)
        i = idx + seq_len
        data = self.data[idx: i]
        target = self.data[idx + 1: i + 1]
        return data, target


class TextFileDataset(TextDataset):
    standard_tokens = []

    @staticmethod
    def download(url: str, path: Path):
        import urllib.request
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        with monit.section("Download") as s:
            def reporthook(count, block_size, total_size):
                s.progress(count * block_size / total_size)

            urllib.request.urlretrieve(url, path, reporthook=reporthook)

    def __init__(self, path: PurePath, tokenizer: Callable, *,
                 url: Optional[str] = None,
                 filter_subset: Optional[int] = None):
        path = Path(path)
        if not path.exists():
            if not url:
                raise FileNotFoundError(str(path))
            else:
                self.download(url, path)

        with monit.section("Load data"):
            text = self.load(path)
            if filter_subset:
                text = text[:filter_subset]
            split = int(len(text) * .9)
            train = text[:split]
            valid = text[split:]

        super().__init__(path, tokenizer, train, valid, '')


def _test_tiny_shakespeare():
    from labml import lab
    _ = TextFileDataset(lab.get_data_path() / 'tiny_shakespeare.txt', lambda x: list(x),
                        url='https://raw.githubusercontent.com/karpathy/char-rnn/master/data/tinyshakespeare/input.txt')


if __name__ == '__main__':
    _test_tiny_shakespeare()



from pathlib import Path
from typing import Optional, Generator

from labml import lab
from labml.logger import inspect


def get_runs() -> Generator[Path, None, None]:
    for exp_path in lab.get_experiments_path().iterdir():
        if exp_path.name.startswith('_'):
            continue
        for run_path in exp_path.iterdir():
            yield run_path


def get_experiments() -> Generator[Path, None, None]:
    for exp_path in lab.get_experiments_path().iterdir():
        if exp_path.name.startswith('_'):
            continue

        yield exp_path


def get_run_by_uuid(run_uuid: str) -> Optional[Path]:
    for exp_path in get_experiments():
        run_path = exp_path / run_uuid
        if run_path.exists():
            return run_path

    return None


def get_runs_by_experiment_path(exp_path: Path) -> Generator[Path, None, None]:
    for run_path in exp_path.iterdir():
        yield run_path


def get_checkpoints(run_path: Path) -> Generator[Path, None, None]:
    checkpoints_path = run_path / "checkpoints"
    if not checkpoints_path.exists():
        return

    for chk_path in checkpoints_path.iterdir():
        yield chk_path


def remove_run(run_path: Path):
    from labml.internal.util import rm_tree
    rm_tree(run_path)


# ===TESTS===
def _test_checkpoints_by_uuid(run_uuid: str):
    run_path = get_run_by_uuid(run_uuid)
    inspect(list(get_checkpoints(run_path)))


def _test_remove_run_by_uuid(run_uuid: str):
    run_path = get_run_by_uuid(run_uuid)
    remove_run(run_path)


if __name__ == '__main__':
    inspect(list(get_runs()))
    # _test_checkpoints_by_uuid('794ca6f21c1f11ebba97acde48001122')
    # _test_remove_run_by_uuid('1443e69c1c2111eb8e8bacde48001122')



import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data

from labml import experiment, tracker
from labml.configs import option
from labml.utils import pytorch as pytorch_utils
from labml_helpers.datasets.mnist import MNISTConfigs
from labml_helpers.device import DeviceConfigs
from labml_helpers.metrics.accuracy import Accuracy
from labml_helpers.module import Module
from labml_helpers.seed import SeedConfigs
from labml_helpers.train_valid import TrainValidConfigs, BatchIndex, hook_model_outputs


class Net(Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, 10)

    def __call__(self, x: torch.Tensor):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        return self.fc2(x)


class Configs(MNISTConfigs, TrainValidConfigs):
    optimizer: torch.optim.Adam
    model: nn.Module
    set_seed = SeedConfigs()
    device: torch.device = DeviceConfigs()
    epochs: int = 10

    is_save_models = True
    model: nn.Module
    inner_iterations = 10

    accuracy_func = Accuracy()
    loss_func = nn.CrossEntropyLoss()

    def init(self):
        tracker.set_queue("loss.*", 20, True)
        tracker.set_scalar("accuracy.*", True)
        hook_model_outputs(self.mode, self.model, 'model')
        self.state_modules = [self.accuracy_func]

    def step(self, batch: any, batch_idx: BatchIndex):
        data, target = batch[0].to(self.device), batch[1].to(self.device)

        if self.mode.is_train:
            tracker.add_global_step(len(data))

        with self.mode.update(is_log_activations=batch_idx.is_last):
            output = self.model(data)

        loss = self.loss_func(output, target)
        self.accuracy_func(output, target)
        tracker.add("loss.", loss)

        if self.mode.is_train:
            loss.backward()

            self.optimizer.step()
            if batch_idx.is_last:
                pytorch_utils.store_model_indicators(self.model)
            self.optimizer.zero_grad()

        tracker.save()


@option(Configs.model)
def model(c: Configs):
    return Net().to(c.device)


@option(Configs.optimizer)
def _optimizer(c: Configs):
    from labml_helpers.optimizer import OptimizerConfigs
    opt_conf = OptimizerConfigs()
    opt_conf.parameters = c.model.parameters()
    return opt_conf


def main():
    conf = Configs()
    experiment.create(name='mnist_labml_helpers')
    experiment.configs(conf, {'optimizer.optimizer': 'Adam'})
    conf.set_seed.set()
    experiment.add_pytorch_models(dict(model=conf.model))
    with experiment.start():
        conf.run()


if __name__ == '__main__':
    main()



"""
---
title: Prioritized Experience Replay Buffer
summary: Annotated implementation of prioritized experience replay using a binary segment tree.
---

# Prioritized Experience Replay Buffer

This implements paper [Prioritized experience replay](https://arxiv.org/abs/1511.05952),
using a binary segment tree.
"""

import random

import numpy as np


class ReplayBuffer:
    """
    ## Buffer for Prioritized Experience Replay

    [Prioritized experience replay](https://arxiv.org/abs/1511.05952)
     samples important transitions more frequently.
    The transitions are prioritized by the Temporal Difference error (td error), $\delta$.

    We sample transition $i$ with probability,
    $$P(i) = \frac{p_i^\alpha}{\sum_k p_k^\alpha}$$
    where $\alpha$ is a hyper-parameter that determines how much
    prioritization is used, with $\alpha = 0$ corresponding to uniform case.
    $p_i$ is the priority.

    We use proportional prioritization $p_i = |\delta_i| + \epsilon$ where
    $\delta_i$ is the temporal difference for transition $i$.

    We correct the bias introduced by prioritized replay using
     importance-sampling (IS) weights
    $$w_i = \bigg(\frac{1}{N} \frac{1}{P(i)}\bigg)^\beta$$ in the loss function.
    This fully compensates when $\beta = 1$.
    We normalize weights by $\frac{1}{\max_i w_i}$ for stability.
    Unbiased nature is most important towards the convergence at end of training.
    Therefore we increase $\beta$ towards end of training.

    ### Binary Segment Tree
    We use a binary segment tree to efficiently calculate
    $\sum_k^i p_k^\alpha$, the cumulative probability,
    which is needed to sample.
    We also use a binary segment tree to find $\min p_i^\alpha$,
    which is needed for $\frac{1}{\max_i w_i}$.
    We can also use a min-heap for this.
    Binary Segment Tree lets us calculate these in $\mathcal{O}(\log n)$
    time, which is way more efficient that the naive $\mathcal{O}(n)$
    approach.

    This is how a binary segment tree works for sum;
    it is similar for minimum.
    Let $x_i$ be the list of $N$ values we want to represent.
    Let $b_{i,j}$ be the $j^{\mathop{th}}$ node of the $i^{\mathop{th}}$ row
     in the binary tree.
    That is two children of node $b_{i,j}$ are $b_{i+1,2j}$ and $b_{i+1,2j + 1}$.

    The leaf nodes on row $D = \left\lceil {1 + \log_2 N} \right\rceil$
     will have values of $x$.
    Every node keeps the sum of the two child nodes.
    That is, the root node keeps the sum of the entire array of values.
    The left and right children of the root node keep
     the sum of the first half of the array and
     the sum of the second half of the array, respectively.
    And so on...

    $$b_{i,j} = \sum_{k = (j -1) * 2^{D - i} + 1}^{j * 2^{D - i}} x_k$$

    Number of nodes in row $i$,
    $$N_i = \left\lceil{\frac{N}{D - i + 1}} \right\rceil$$
    This is equal to the sum of nodes in all rows above $i$.
    So we can use a single array $a$ to store the tree, where,
    $$b_{i,j} \rightarrow a_{N_i + j}$$

    Then child nodes of $a_i$ are $a_{2i}$ and $a_{2i + 1}$.
    That is,
    $$a_i = a_{2i} + a_{2i + 1}$$

    This way of maintaining binary trees is very easy to program.
    *Note that we are indexing starting from 1*.

    We use the same structure to compute the minimum.
    """

    def __init__(self, capacity, alpha):
        """
        ### Initialize
        """
        # We use a power of $2$ for capacity because it simplifies the code and debugging
        self.capacity = capacity
        # $\alpha$
        self.alpha = alpha

        # Maintain segment binary trees to take sum and find minimum over a range
        self.priority_sum = [0 for _ in range(2 * self.capacity)]
        self.priority_min = [float('inf') for _ in range(2 * self.capacity)]

        # Current max priority, $p$, to be assigned to new transitions
        self.max_priority = 1.

        # Arrays for buffer
        self.data = {
            'obs': np.zeros(shape=(capacity, 4, 84, 84), dtype=np.uint8),
            'action': np.zeros(shape=capacity, dtype=np.int32),
            'reward': np.zeros(shape=capacity, dtype=np.float32),
            'next_obs': np.zeros(shape=(capacity, 4, 84, 84), dtype=np.uint8),
            'done': np.zeros(shape=capacity, dtype=np.bool)
        }
        # We use cyclic buffers to store data, and `next_idx` keeps the index of the next empty
        # slot
        self.next_idx = 0

        # Size of the buffer
        self.size = 0

    def add(self, obs, action, reward, next_obs, done):
        """
        ### Add sample to queue
        """

        # Get next available slot
        idx = self.next_idx

        # store in the queue
        self.data['obs'][idx] = obs
        self.data['action'][idx] = action
        self.data['reward'][idx] = reward
        self.data['next_obs'][idx] = next_obs
        self.data['done'][idx] = done

        # Increment next available slot
        self.next_idx = (idx + 1) % self.capacity
        # Calculate the size
        self.size = min(self.capacity, self.size + 1)

        # $p_i^\alpha$, new samples get `max_priority`
        priority_alpha = self.max_priority ** self.alpha
        # Update the two segment trees for sum and minimum
        self._set_priority_min(idx, priority_alpha)
        self._set_priority_sum(idx, priority_alpha)

    def _set_priority_min(self, idx, priority_alpha):
        """
        #### Set priority in binary segment tree for minimum
        """

        # Leaf of the binary tree
        idx += self.capacity
        self.priority_min[idx] = priority_alpha

        # Update tree, by traversing along ancestors.
        # Continue until the root of the tree.
        while idx >= 2:
            # Get the index of the parent node
            idx //= 2
            # Value of the parent node is the minimum of it's two children
            self.priority_min[idx] = min(self.priority_min[2 * idx], self.priority_min[2 * idx + 1])

    def _set_priority_sum(self, idx, priority):
        """
        #### Set priority in binary segment tree for sum
        """

        # Leaf of the binary tree
        idx += self.capacity
        # Set the priority at the leaf
        self.priority_sum[idx] = priority

        # Update tree, by traversing along ancestors.
        # Continue until the root of the tree.
        while idx >= 2:
            # Get the index of the parent node
            idx //= 2
            # Value of the parent node is the sum of it's two children
            self.priority_sum[idx] = self.priority_sum[2 * idx] + self.priority_sum[2 * idx + 1]

    def _sum(self):
        """
        #### $\sum_k p_k^\alpha$
        """

        # The root node keeps the sum of all values
        return self.priority_sum[1]

    def _min(self):
        """
        #### $\min_k p_k^\alpha$
        """

        # The root node keeps the minimum of all values
        return self.priority_min[1]

    def find_prefix_sum_idx(self, prefix_sum):
        """
        #### Find largest $i$ such that $\sum_{k=1}^{i} p_k^\alpha  \le P$
        """

        # Start from the root
        idx = 1
        while idx < self.capacity:
            # If the sum of the left branch is higher than required sum
            if self.priority_sum[idx * 2] > prefix_sum:
                # Go to left branch of the tree
                idx = 2 * idx
            else:
                # Otherwise go to right branch and reduce the sum of left
                #  branch from required sum
                prefix_sum -= self.priority_sum[idx * 2]
                idx = 2 * idx + 1

        # We are at the leaf node. Subtract the capacity by the index in the tree
        # to get the index of actual value
        return idx - self.capacity

    def sample(self, batch_size, beta):
        """
        ### Sample from buffer
        """

        # Initialize samples
        samples = {
            'weights': np.zeros(shape=batch_size, dtype=np.float32),
            'indexes': np.zeros(shape=batch_size, dtype=np.int32)
        }

        # Get sample indexes
        for i in range(batch_size):
            p = random.random() * self._sum()
            idx = self.find_prefix_sum_idx(p)
            samples['indexes'][i] = idx

        # $\min_i P(i) = \frac{\min_i p_i^\alpha}{\sum_k p_k^\alpha}$
        prob_min = self._min() / self._sum()
        # $\max_i w_i = \bigg(\frac{1}{N} \frac{1}{\min_i P(i)}\bigg)^\beta$
        max_weight = (prob_min * self.size) ** (-beta)

        for i in range(batch_size):
            idx = samples['indexes'][i]
            # $P(i) = \frac{p_i^\alpha}{\sum_k p_k^\alpha}$
            prob = self.priority_sum[idx + self.capacity] / self._sum()
            # $w_i = \bigg(\frac{1}{N} \frac{1}{P(i)}\bigg)^\beta$
            weight = (prob * self.size) ** (-beta)
            # Normalize by $\frac{1}{\max_i w_i}$,
            #  which also cancels off the $\frac{1}{N}$ term
            samples['weights'][i] = weight / max_weight

        # Get samples data
        for k, v in self.data.items():
            samples[k] = v[samples['indexes']]

        return samples

    def update_priorities(self, indexes, priorities):
        """
        ### Update priorities
        """

        for idx, priority in zip(indexes, priorities):
            # Set current max priority
            self.max_priority = max(self.max_priority, priority)

            # Calculate $p_i^\alpha$
            priority_alpha = priority ** self.alpha
            # Update the trees
            self._set_priority_min(idx, priority_alpha)
            self._set_priority_sum(idx, priority_alpha)

    def is_full(self):
        """
        ### Whether the buffer is full
        """
        return self.capacity == self.size



"""
---
title: Adam optimizer with warm-up
summary: A simple PyTorch implementation/tutorial of Adam optimizer with warm-up.
---

# Adam Optimizer with Warmup

This extends [AMSGrad optimizer](adam.html) and adds a warmup stage.
"""

from typing import Dict

from labml_nn.optimizers import WeightDecay
from labml_nn.optimizers.amsgrad import AMSGrad


class AdamWarmup(AMSGrad):
    """
    ## Adam Optimizer with Warmup

    This class extends from Adam optimizer defined in [`adam.py`](adam.html).
    """
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-16,
                 weight_decay: WeightDecay = WeightDecay(),
                 optimized_update: bool = True,
                 amsgrad=False, warmup=0, defaults=None):
        """
        ### Initialize the optimizer

        * `params` is the list of parameters
        * `lr` is the learning rate $\alpha$
        * `betas` is a tuple of ($\beta_1$, $\beta_2$)
        * `eps` is $\hat{\epsilon}$ or $\epsilon$ based on `optimized_update`
        * `weight_decay` is an instance of class `WeightDecay` defined in [`__init__.py`](index.html)
        * 'optimized_update' is a flag whether to optimize the bias correction of the second moment
          by doing it after adding $\epsilon$
        * `amsgrad` is a flag indicating whether to use AMSGrad or fallback to plain Adam
        * `warmup` number of warmup steps
        * `defaults` is a dictionary of default for group values.
         This is useful when you want to extend the class `AdamWarmup`.
        """

        defaults = {} if defaults is None else defaults
        defaults.update(dict(warmup=warmup))
        super().__init__(params, lr, betas, eps, weight_decay, optimized_update, amsgrad, defaults)

    def get_lr(self, state: Dict[str, any], group: Dict[str, any]):
        """
        ### Get learning-rate

        $$\alpha \min \bigg(1, \frac{t}{w}\bigg)$$
        where $w$ is the number of warmup steps.
        """
        # If we are in warmup stage
        if group['warmup'] > state['step']:
            # A linearly increasing learning rate from $0$ to $\alpha$
            return 1e-8 + state['step'] * group['lr'] / group['warmup']
        else:
            # Constant learning rate $\alpha$
            return group['lr']



from abc import ABC
from collections import deque
from typing import Dict

import numpy as np

from . import Indicator
from labml.internal.util.values import to_numpy


class NumericIndicator(Indicator, ABC):
    def get_mean(self) -> float:
        raise NotImplementedError()

    def get_histogram(self):
        raise NotImplementedError()

    @property
    def mean_key(self):
        return f'{self.name}.mean'


class Queue(NumericIndicator):
    def __init__(self, name: str, queue_size=10, is_print=False):
        super().__init__(name=name, is_print=is_print)
        self.queue_size = queue_size
        self._values = deque(maxlen=queue_size)
        self._is_empty = True

    def collect_value(self, value):
        self._is_empty = False
        self._values.append(to_numpy(value).ravel())

    def to_dict(self) -> Dict:
        res = super().to_dict().copy()
        res.update({'queue_size': self._values.maxlen})
        return res

    def is_empty(self) -> bool:
        return len(self._values) == 0 or self._is_empty

    def clear(self):
        self._is_empty = True

    def get_mean(self) -> float:
        return float(np.mean(self._values))

    def get_histogram(self):
        return self._values

    @property
    def mean_key(self):
        return f'{self.name}.mean'

    def copy(self, key: str):
        return Queue(key, queue_size=self.queue_size, is_print=self.is_print)

    def equals(self, value: any) -> bool:
        if not super().equals(value):
            return False
        return value.queue_size == self.queue_size


class _Collection(NumericIndicator, ABC):
    def __init__(self, name: str, is_print: bool):
        super().__init__(name=name, is_print=is_print)
        self._values = []

    def _merge(self):
        if len(self._values) == 0:
            return []
        elif len(self._values) == 1:
            return self._values[0]
        else:
            merged = np.concatenate(self._values, axis=0)
            self._values = [merged]

            return merged

    def collect_value(self, value):
        self._values.append(to_numpy(value).ravel())

    def clear(self):
        self._values = []

    def is_empty(self) -> bool:
        return len(self._values) == 0

    def get_mean(self) -> float:
        return float(np.mean(self._merge()))

    def get_histogram(self):
        return self._merge()


class Histogram(_Collection):
    def copy(self, key: str):
        return Histogram(key, is_print=self.is_print)


class Scalar(_Collection):
    def get_histogram(self):
        return None

    def copy(self, key: str):
        return Scalar(key, is_print=self.is_print)

    @property
    def mean_key(self):
        return f'{self.name}'



from labml.internal.logger.destinations import Destination
from labml.internal.util import is_ipynb, is_ipynb_pycharm


def create_destination() -> Destination:
    if is_ipynb():
        if is_ipynb_pycharm():
            from labml.internal.logger.destinations.ipynb_pycharm import IpynbPyCharmDestination
            return IpynbPyCharmDestination()
        else:
            from labml.internal.logger.destinations.ipynb import IpynbDestination
            return IpynbDestination()
    else:
        from labml.internal.logger.destinations.console import ConsoleDestination
        return ConsoleDestination()



import setuptools

with open("readme.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name='labml_nn',
    version='0.4.64',
    author="Varuna Jayasiri, Nipun Wijerathne",
    author_email="vpjayasiri@gmail.com, hnipun@gmail.com",
    description="A collection of PyTorch implementations of neural network architectures and layers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lab-ml/labml_nn",
    project_urls={
        'Documentation': 'https://lab-ml.com/'
    },
    packages=setuptools.find_packages(exclude=('test',
                                               'test.*')),
    install_requires=['labml>=0.4.62',
                      'labml_helpers>=0.4.62',
                      'torch',
                      'einops',
                      'numpy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='machine learning',
)



import time

from numpy.random import random

from labml import tracker, experiment

conf = {'batch_size': 20}


def train(n: int):
    return 0.999 ** n + random() / 10, 1 - .999 ** n + random() / 10


with experiment.record(name='sample', exp_conf=conf):
    for i in range(100000):
        time.sleep(0.2)
        loss, accuracy = train(i)
        tracker.save(i, {'loss': loss, 'accuracy': accuracy})



"""
[![PiPy Version](https://badge.fury.io/py/labml-nn.svg)](https://badge.fury.io/py/labml-nn)
[![PiPy Downloads](https://pepy.tech/badge/labml-nn)](https://pepy.tech/project/labml-nn)

# [LabML Neural Networks](https://lab-ml.com/labml_nn/index.html)

This is a collection of simple PyTorch implementations of
neural networks and related algorithms.
[These implementations](https://github.com/lab-ml/nn) are documented with explanations,
and the [website](https://lab-ml.com/labml_nn/index.html)
renders these as side-by-side formatted notes.
We believe these would help you understand these algorithms better.

We are actively maintaining this repo and adding new
implementations.

## Modules

#### ✨ [Transformers](https://lab-ml.com/labml_nn/transformers)

[Transformers module](https://lab-ml.com/labml_nn/transformers)
contains implementations for
[multi-headed attention](https://lab-ml.com/labml_nn/transformers/mha.html)
and
[relative multi-headed attention](https://lab-ml.com/labml_nn/transformers/relative_mha.html).

* [kNN-LM: Generalization through Memorization](https://lab-ml.com/labml_nn/transformers/knn)

#### ✨ [Recurrent Highway Networks](https://lab-ml.com/labml_nn/recurrent_highway_networks)

#### ✨ [LSTM](https://lab-ml.com/labml_nn/lstm)

#### ✨ [Capsule Networks](https://lab-ml.com/labml_nn/capsule_networks/)

#### ✨ [Generative Adversarial Networks](https://lab-ml.com/labml_nn/gan/)
* [GAN with a multi-layer perceptron](https://lab-ml.com/labml_nn/gan/simple_mnist_experiment.html)
* [GAN with deep convolutional network](https://lab-ml.com/labml_nn/gan/dcgan.html)
* [Cycle GAN](https://lab-ml.com/labml_nn/gan/cycle_gan.html)

#### ✨ [Sketch RNN](https://lab-ml.com/labml_nn/sketch_rnn/)

#### ✨ [Reinforcement Learning](https://lab-ml.com/labml_nn/rl/)
* [Proximal Policy Optimization](https://lab-ml.com/labml_nn/rl/ppo/) with
 [Generalized Advantage Estimation](https://lab-ml.com/labml_nn/rl/ppo/gae.html)
* [Deep Q Networks](https://lab-ml.com/labml_nn/rl/dqn/) with
 with [Dueling Network](https://lab-ml.com/labml_nn/rl/dqn/model.html),
 [Prioritized Replay](https://lab-ml.com/labml_nn/rl/dqn/replay_buffer.html)
 and Double Q Network.

#### ✨ [Optimizers](https://lab-ml.com/labml_nn/optimizers/)
* [Adam](https://lab-ml.com/labml_nn/optimizers/adam.html)
* [AMSGrad](https://lab-ml.com/labml_nn/optimizers/amsgrad.html)
* [Adam Optimizer with warmup](https://lab-ml.com/labml_nn/optimizers/adam_warmup.html)
* [Noam Optimizer](https://lab-ml.com/labml_nn/optimizers/noam.html)
* [Rectified Adam Optimizer](https://lab-ml.com/labml_nn/optimizers/radam.html)
* [AdaBelief Optimizer](https://lab-ml.com/labml_nn/optimizers/ada_belief.html)

### Installation

```bash
pip install labml_nn
```

### Citing LabML

If you use LabML for academic research, please cite the library using the following BibTeX entry.

```bibtex
@misc{labml,
 author = {Varuna Jayasiri, Nipun Wijerathne},
 title = {LabML: A library to organize machine learning experiments},
 year = {2020},
 url = {https://lab-ml.com/},
}
```
"""



import typing
from typing import Optional, List, Union, Tuple

from .iterator import Iterator
from .loop import Loop
from .sections import Section, OuterSection
from ..logger import logger_singleton as logger
from ..logger.types import LogPart
from ..tracker import tracker_singleton as tracker
from labml.internal.util.colors import StyleCode
from ...logger import Text


class Monitor:
    __loop_indicators: List[Union[str, Tuple[str, Optional[StyleCode]]]]
    __is_looping: bool

    def __init__(self):
        self.__loop: Optional[Loop] = None
        self.__sections: List[Section] = []
        self.__is_looping = False
        self.__loop_indicators = []
        self.__is_silent = False

    def clear(self):
        self.__loop: Optional[Loop] = None
        self.__sections: List[Section] = []
        self.__is_looping = False
        self.__loop_indicators = []

    def silent(self, is_silent: bool = True):
        self.__is_silent = is_silent

    def iterate(self, name, iterable: Union[typing.Iterable, typing.Sized, int],
                total_steps: Optional[int], *,
                is_silent: bool,
                is_children_silent: bool,
                is_timed: bool,
                section: Optional[Section]):
        return Iterator(logger=self,
                        name=name,
                        iterable=iterable,
                        is_silent=is_silent,
                        is_timed=is_timed,
                        total_steps=total_steps,
                        is_children_silent=is_children_silent,
                        is_enumerate=False,
                        section=section)

    def enum(self, name, iterable: typing.Sized, *,
             is_silent: bool,
             is_children_silent: bool,
             is_timed: bool,
             section: Optional[Section]):
        return Iterator(logger=self,
                        name=name,
                        iterable=iterable,
                        is_silent=is_silent,
                        is_timed=is_timed,
                        total_steps=None,
                        is_children_silent=is_children_silent,
                        is_enumerate=True,
                        section=section)

    def section(self, name, *,
                is_silent: bool,
                is_timed: bool,
                is_partial: bool,
                is_new_line: bool,
                is_children_silent: bool,
                total_steps: float) -> Section:
        if self.__is_looping:
            if len(self.__sections) != 0:
                is_silent = True

            section = self.__loop.get_section(name=name,
                                              is_silent=is_silent,
                                              is_timed=is_timed,
                                              is_partial=is_partial,
                                              total_steps=total_steps,
                                              parents=[s.name for s in self.__sections])
            self.__sections.append(section)
        else:
            if len(self.__sections) > 0:
                if self.__sections[-1].is_silent or self.__sections[-1].is_children_silent:
                    is_silent = True
                    is_children_silent = True
            self.__sections.append(OuterSection(monitor=self,
                                                name=name,
                                                is_silent=is_silent,
                                                is_timed=is_timed,
                                                is_partial=is_partial,
                                                is_new_line=is_new_line,
                                                is_children_silent=is_children_silent,
                                                total_steps=total_steps,
                                                level=len(self.__sections)))

        return self.__sections[-1]

    def progress(self, steps: float):
        if len(self.__sections) == 0:
            raise RuntimeError("You must be within a section to report progress")

        if self.__sections[-1].progress(steps):
            self.__log_line()

    def set_successful(self, is_successful=True):
        if len(self.__sections) == 0:
            raise RuntimeError("You must be within a section to report success")

        self.__sections[-1].is_successful = is_successful
        self.__log_line()

    def loop(self, iterator_: typing.Collection, *,
             is_track: bool,
             is_print_iteration_time: bool):
        if len(self.__sections) != 0:
            raise RuntimeError("Cannot start a loop within a section")

        self.__loop = Loop(iterator=iterator_,
                           monitor=self,
                           is_track=is_track,
                           is_print_iteration_time=is_print_iteration_time)
        return self.__loop

    def start_loop(self):
        self.__is_looping = True
        tracker().start_loop(self.set_looping_indicators)

    def finish_loop(self):
        if len(self.__sections) != 0:
            raise RuntimeError("Cannot be within a section when finishing the loop")
        tracker().finish_loop()
        self.__loop = None
        self.__is_looping = False

    def section_enter(self, section):
        if len(self.__sections) == 0:
            raise RuntimeError("Entering a section without creating a section.\n"
                               "Always use logger.section to create a section")

        if section is not self.__sections[-1]:
            raise RuntimeError("Entering a section other than the one last_created\n"
                               "Always user with logger.section(...):")

        if len(self.__sections) > 1 and not self.__sections[-2].is_parented:
            self.__sections[-2].make_parent()
            if not self.__is_silent and not self.__sections[-1].is_silent:
                logger().log([])

        self.__log_line()

    def __log_looping_line(self):
        parts = [(f"{tracker().global_step :8,}:  ", Text.highlight)]
        parts += self.__loop.log_sections()
        parts += self.__loop_indicators
        parts += self.__loop.log_progress()

        if not self.__is_silent:
            logger().log(parts, is_new_line=False)

    def __log_line(self):
        if self.__is_looping:
            self.__log_looping_line()
            return

        if len(self.__sections) == 0:
            return

        parts = self.__sections[-1].log()
        if parts is None:
            return

        if not self.__is_silent:
            logger().log(parts, is_new_line=False)

    def set_looping_indicators(self, indicators: List[LogPart]):
        self.__loop_indicators = indicators
        self.__log_looping_line()

    def section_exit(self, section):
        if len(self.__sections) == 0:
            raise RuntimeError("Impossible")

        if section is not self.__sections[-1]:
            raise RuntimeError("Impossible")

        self.__log_line()
        self.__sections.pop(-1)


_internal: Optional[Monitor] = None


def monitor_singleton() -> Monitor:
    global _internal
    if _internal is None:
        _internal = Monitor()

    return _internal



import types
import warnings
from collections import OrderedDict
from typing import Dict, List, Callable, Union, Tuple, Optional, Type, Set, Iterable

from .config_function import ConfigFunction
from .config_item import ConfigItem
from .eval_function import EvalFunction
from .utils import Value
from ... import monit


def _is_class_method(func: Callable):
    if not isinstance(func, types.FunctionType):
        return False
    # if not callable(func):
    #     return False

    import inspect

    spec: inspect.Signature = inspect.signature(func)
    params: List[inspect.Parameter] = list(spec.parameters.values())
    if len(params) == 0:
        raise RuntimeError('Can only have methods in a config class', func)
    p = params[0]
    if p.kind != p.POSITIONAL_OR_KEYWORD:
        raise RuntimeError('Can only have methods in a config class', func)
    if p.name != 'self':
        raise RuntimeError('Can only have methods in a config class', func)

    return True


RESERVED_CLASS = {'calc', 'list', 'set_hyperparams', 'set_meta', 'aggregate', 'calc_wrap'}
RESERVED_INSTANCE = {'_to_json', '_reset_explicitly_specified', '_set_update_callback', '_set_values', '_get_type'}

_STANDARD_TYPES = {int, str, bool, float, Dict, List}


def _is_valid(key):
    if key.startswith('_'):
        return False

    if key in RESERVED_CLASS or key in RESERVED_INSTANCE:
        return False

    return True


def _get_base_classes(class_: Type['Configs']) -> List[Type['Configs']]:
    classes = [class_]
    level = [class_]
    next_level = []

    while len(level) > 0:
        for c in level:
            for b in c.__bases__:
                if b == object:
                    continue
                next_level.append(b)
        classes += next_level
        level = next_level
        next_level = []

    classes.reverse()

    unique_classes = []
    hashes: Set[int] = set()
    for c in classes:
        if hash(c) not in hashes:
            unique_classes.append(c)
        hashes.add(hash(c))

    return unique_classes


class PropertyKeys:
    calculators = '_calculators'
    evaluators = '_evaluators'
    hyperparams = '_hyperparams'
    aggregates = '_aggregates'
    meta = '_meta'


class Configs:
    _calculators: Dict[str, List[ConfigFunction]] = {}
    _evaluators: Dict[str, List[EvalFunction]] = {}
    _hyperparams: Dict[str, bool]
    _aggregates: Dict[str, Dict[str, Dict[ConfigItem, any]]]
    _meta: Dict[str, bool]

    __config_items: Dict[str, ConfigItem]
    __options: Dict[str, Dict[str, ConfigFunction]]
    __evals: Dict[str, Dict[str, EvalFunction]]
    __types: Dict[str, Type]
    __explicitly_specified: Set[str]
    __hyperparams: Dict[str, bool]
    __meta: Dict[str, bool]
    __aggregates: Dict[str, Dict[str, Dict[str, any]]]
    __aggregates_options: Dict[str, Set[str]]
    __secondary_values: Dict[str, Dict[str, any]]

    __values: Dict[str, any]
    __values_override: Dict[str, any]
    __cached: Dict[str, any]
    __cached_configs: Dict[str, 'Configs']

    __order: Dict[str, int]
    __n_calculated: int

    __update_callback: Optional[Callable]

    def __init__(self, *, _primary: str = None):
        self._primary = _primary
        self.__values = {}
        self.__values_override = {}
        self.__cached = {}
        self.__cached_configs = {}

        classes = _get_base_classes(type(self))

        self.__types = {}
        self.__options = {}
        self.__evals = {}
        self.__config_items = {}
        self.__explicitly_specified = set()
        self.__hyperparams = {}
        self.__meta = {}
        self.__aggregates = {}
        self.__aggregates_options = {}
        self.__secondary_values = {}

        self.__order = {}
        self.__n_calculated = 0

        self.__update_callback = None

        self.__collect_config_items(classes)
        self.__collect_calculator(classes)
        self.__collect_evaluator(classes)
        self.__collect_aggregates(classes)

        for c in classes:
            if PropertyKeys.hyperparams in c.__dict__:
                for k, is_hyperparam in c.__dict__[PropertyKeys.hyperparams].items():
                    self.__hyperparams[k] = is_hyperparam

        for c in classes:
            if PropertyKeys.meta in c.__dict__:
                for k, is_meta in c.__dict__[PropertyKeys.meta].items():
                    self.__meta[k] = is_meta

    def __collect_config_items(self, classes: List[Type['Configs']]):
        for c in classes:
            for k, v in c.__dict__.items():
                if PropertyKeys.evaluators in c.__dict__ and k in c.__dict__[PropertyKeys.evaluators]:
                    continue
                if not _is_valid(k):
                    continue

                if v.has_value:
                    self.__values[k] = v.value

                if k in self.__config_items:
                    self.__config_items[k].update(v)
                else:
                    self.__config_items[k] = v

                if k not in self.__types:
                    self.__types[k] = v.annotation

    def __collect_calculator(self, classes: List[Type['Configs']]):
        for c in classes:
            if PropertyKeys.calculators not in c.__dict__:
                continue
            for k, calculators in c.__dict__[PropertyKeys.calculators].items():
                if k not in self.__types:
                    raise RuntimeError(f"{k} calculator is present but the config declaration is missing")
                for v in calculators:
                    if k not in self.__options:
                        self.__options[k] = OrderedDict()
                    if v.option_name in self.__options[k]:
                        if v != self.__options[k][v.option_name]:
                            warnings.warn(f"Overriding option for {k}: {v.option_name}", Warning, stacklevel=5)

                    self.__options[k][v.option_name] = v

    def __collect_evaluator(self, classes: List[Type['Configs']]):
        for c in classes:
            if PropertyKeys.evaluators not in c.__dict__:
                continue
            for k, evaluators in c.__dict__[PropertyKeys.evaluators].items():
                for v in evaluators:
                    if k not in self.__evals:
                        self.__evals[k] = OrderedDict()

                    self.__evals[k]['default'] = v

    def __collect_aggregates(self, classes: List[Type['Configs']]):
        for c in classes:
            if PropertyKeys.aggregates not in c.__dict__:
                continue
            for key, aggregates in c.__dict__[PropertyKeys.aggregates].items():
                for option, pairs in aggregates.items():
                    if key not in self.__aggregates_options:
                        self.__aggregates_options[key] = set()
                    self.__aggregates_options[key].add(option)
                    for name, value in pairs.items():
                        if name not in self.__aggregates:
                            self.__aggregates[name] = {}
                        if key not in self.__aggregates[name]:
                            self.__aggregates[name][key] = {}
                        self.__aggregates[name][key][option] = value

    def __dir__(self) -> Iterable[str]:
        return [k for k in self.__types]

    def __setattr__(self, key, value):
        if key.startswith('_'):
            self.__dict__[key] = value
        elif key in self.__types:
            if key in self.__cached:
                raise ValueError(f"Cannot set {self.__class__.__name__}:{key} after it was accessed")
            self.__values[key] = value
        else:
            raise AttributeError(f"{self.__class__.__name__} has no annotation for attribute '{key}'")

    def __getattribute__(self, item):
        if item.startswith('_'):
            return object.__getattribute__(self, item)
        if item in RESERVED_INSTANCE:
            return object.__getattribute__(self, item)
        if item in self.__evals:
            return object.__getattribute__(self, item)

        if item not in self.__cached:
            self.__calculate(item)
            if self.__update_callback is not None:
                self.__update_callback()
        return self.__cached[item]

    def __get_value_aggregate(self, item):
        if item not in self.__aggregates:
            return None, False

        for k, options in self.__aggregates[item].items():
            v, has = self.__get_value(k)
            if has and v in options:
                return options[v], True

        return None, False

    def __get_value_direct(self, item):
        if item in self.__values_override:
            return self.__values_override[item], True
        elif item in self.__values:
            return self.__values[item], True
        else:
            return None, False

    def __get_value(self, item):
        if item in self.__cached:
            return self.__cached[item]

        value, has = self.__get_value_direct(item)

        if has and value != '__aggregate__':
            return value, has

        value, has = self.__get_value_aggregate(item)
        if has:
            return value, has

        if item in self.__options:
            return next(iter(self.__options[item].keys())), True

        return None, False

    def __calculate(self, item):
        self.__order[item] = self.__n_calculated
        self.__n_calculated += 1

        value, has = self.__get_value(item)

        if has:
            pass
        elif item not in self.__types:
            raise AttributeError(f"{self.__class__.__name__} has no attribute `{item}`")
        elif type(self.__types[item]) == type and self.__types[item] not in _STANDARD_TYPES:
            config_function = ConfigFunction(self.__types[item],
                                             config_names=self.__config_items[item],
                                             option_name='from_type')

            self.__options[item] = OrderedDict()
            self.__options[item]['from_type'] = config_function
            self.__values[item] = 'from_type'
            value = 'from_type'
        else:
            raise AttributeError(f"{self.__class__.__name__} cannot calculate config `{item}`")

        if item in self.__options and value in self.__options[item]:
            func = self.__options[item][value]
            with monit.section(f'Prepare {item}'):
                value = func(self)

        if isinstance(value, Configs):
            self.__cached_configs[item] = value
            value._reset_explicitly_specified()
            if item in self.__secondary_values:
                s_values = self.__secondary_values[item]
                del self.__secondary_values[item]
                value._set_values(s_values)

            primary = value.__dict__.get('_primary', None)
            if primary is not None:
                value = getattr(value, primary)

        self.__cached[item] = value

    def __init_subclass__(cls, **kwargs):
        configs = {}

        for k, v in cls.__annotations__.items():
            if not _is_valid(k):
                continue

            configs[k] = ConfigItem(key=k,
                                    configs_class=cls,
                                    has_annotation=True, annotation=v,
                                    has_value=k in cls.__dict__,
                                    value=cls.__dict__.get(k, None))

        evals = []

        for k, v in cls.__dict__.items():
            if not _is_valid(k):
                continue

            if _is_class_method(v):
                evals.append((k, v))
                continue

            configs[k] = ConfigItem(key=k,
                                    configs_class=cls,
                                    has_annotation=k in cls.__annotations__,
                                    annotation=cls.__annotations__.get(k, None),
                                    has_value=True, value=v)

        for e in evals:
            cls._add_eval_function(e[1], e[0])

        for k, v in configs.items():
            setattr(cls, k, v)

    @classmethod
    def _add_config_function(cls,
                             func: Callable,
                             name: Union[ConfigItem, List[ConfigItem]],
                             option: Optional[str],
                             pass_params: Optional[List[ConfigItem]]):
        if PropertyKeys.calculators not in cls.__dict__:
            cls._calculators = {}

        calc = ConfigFunction(func,
                              config_names=name,
                              option_name=option,
                              pass_params=pass_params)
        if type(calc.config_names) == str:
            config_names = [calc.config_names]
        else:
            config_names = calc.config_names

        for n in config_names:
            if n not in cls._calculators:
                cls._calculators[n] = []
            cls._calculators[n].append(calc)

    @classmethod
    def _add_eval_function(cls,
                           func: Callable,
                           name: str):
        if PropertyKeys.evaluators not in cls.__dict__:
            cls._evaluators = {}

        calc = EvalFunction(func,
                            config_name=name)

        if name not in cls._evaluators:
            cls._evaluators[name] = []
        cls._evaluators[name].append(calc)

    @classmethod
    def calc_wrap(cls, func: Callable,
                  name: Union[ConfigItem, List[ConfigItem]],
                  option_name: Optional[str] = None,
                  pass_params: Optional[List[ConfigItem]] = None):
        cls._add_config_function(func, name, option_name, pass_params)

        return func

    @classmethod
    def calc(cls,
             name: Union[ConfigItem, List[ConfigItem]] = None,
             option_name: Optional[str] = None,
             pass_params: Optional[List[ConfigItem]] = None):

        def wrapper(func: Callable):
            cls._add_config_function(func, name, option_name, pass_params)

            return func

        return wrapper

    @classmethod
    def set_hyperparams(cls, *args: ConfigItem, is_hyperparam=True):
        if PropertyKeys.hyperparams not in cls.__dict__:
            cls._hyperparams = {}

        for h in args:
            cls._hyperparams[h.key] = is_hyperparam

    @classmethod
    def set_meta(cls, *args: ConfigItem, is_meta=True):
        if PropertyKeys.meta not in cls.__dict__:
            cls._meta = {}

        for h in args:
            cls._meta[h.key] = is_meta

    @classmethod
    def aggregate(cls, name: ConfigItem, option: str,
                  *args: Tuple[ConfigItem, any]):
        assert args

        if PropertyKeys.aggregates not in cls.__dict__:
            cls._aggregates = {}

        if name.key not in cls._aggregates:
            cls._aggregates[name.key] = {}

        pairs = {p[0].key: p[1] for p in args}
        cls._aggregates[name.key][option] = pairs

    def _set_values(self, values: Dict[str, any]):
        sub_modules = set()
        for k, v in values.items():
            if not _is_valid(k):
                raise KeyError(f"Invalid attribute name '{k}'")
            if k in self.__types:
                if k in self.__cached:
                    raise ValueError(f"Cannot set {self.__class__.__name__}:{k} after it was accessed")

                self.__explicitly_specified.add(k)

                self.__values_override[k] = v
            else:
                tk, *_ = k.split('.')
                sub_modules.add(tk)
                if tk not in self.__types:
                    raise KeyError(f"Invalid attribute name '{tk}' (taken from '{k}')")

                if tk not in self.__secondary_values:
                    self.__secondary_values[tk] = {}
                self.__secondary_values[tk][k[len(tk) + 1:]] = v

        for k in sub_modules:
            if k in self.__cached_configs:
                s_values = self.__secondary_values[k]
                del self.__secondary_values[k]
                self.__cached_configs[k]._set_values(s_values)

    def __get_options_list(self, key: str):
        opts = list(self.__options.get(key, {}).keys())
        if not opts:
            opts = list((self.__aggregates_options.get(key, set())))

        return opts

    def _to_json(self):
        configs = {}
        for k, v in self.__types.items():
            configs[k] = {
                'name': k,
                'type': str(v),
                'value': Value.to_yaml_truncated(self.__values.get(k, None)),
                'order': self.__order.get(k, -1),
                'options': self.__get_options_list(k),
                'computed': Value.to_yaml_truncated(self.__cached.get(k, None)),
                'is_hyperparam': self.__hyperparams.get(k, None),
                'is_meta': self.__meta.get(k, None),
                'is_explicitly_specified': (k in self.__explicitly_specified)
            }

        for k, c in self.__cached_configs.items():
            sub_configs = c._to_json()
            for sk, v in sub_configs.items():
                configs[f"{k}.{sk}"] = v

        return configs

    def _reset_explicitly_specified(self):
        self.__explicitly_specified = set()
        for k, v in self.__cached_configs:
            v._reset_explicitly_specified()

    def _set_update_callback(self, update_callback: Callable):
        self.__update_callback = update_callback

    def _get_type(self, item: str) -> Type:
        return self.__types[item]



from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from labml import lab
from labml.configs import BaseConfigs


def _dataset(is_train, transform):
    return datasets.CIFAR10(str(lab.get_data_path()),
                            train=is_train,
                            download=True,
                            transform=transform)


class CIFAR10Configs(BaseConfigs):
    dataset_name: str = 'CIFAR10'
    dataset_transforms: transforms.Compose
    train_dataset: datasets.CIFAR10
    valid_dataset: datasets.CIFAR10

    train_loader: DataLoader
    valid_loader: DataLoader

    train_batch_size: int = 64
    valid_batch_size: int = 1024

    train_loader_shuffle: bool = True
    valid_loader_shuffle: bool = False


@CIFAR10Configs.calc(CIFAR10Configs.dataset_transforms)
def cifar10_transforms():
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])


@CIFAR10Configs.calc(CIFAR10Configs.train_dataset)
def cifar10_train_dataset(c: CIFAR10Configs):
    return _dataset(True, c.dataset_transforms)


@CIFAR10Configs.calc(CIFAR10Configs.valid_dataset)
def cifar10_valid_dataset(c: CIFAR10Configs):
    return _dataset(False, c.dataset_transforms)


@CIFAR10Configs.calc(CIFAR10Configs.train_loader)
def cifar10_train_loader(c: CIFAR10Configs):
    return DataLoader(c.train_dataset,
                      batch_size=c.train_batch_size,
                      shuffle=c.train_loader_shuffle)


@CIFAR10Configs.calc(CIFAR10Configs.valid_loader)
def cifar10_valid_loader(c: CIFAR10Configs):
    return DataLoader(c.valid_dataset,
                      batch_size=c.valid_batch_size,
                      shuffle=c.valid_loader_shuffle)


CIFAR10Configs.aggregate(CIFAR10Configs.dataset_name, 'CIFAR10',
                       (CIFAR10Configs.dataset_transforms, 'cifar10_transforms'),
                       (CIFAR10Configs.train_dataset, 'cifar10_train_dataset'),
                       (CIFAR10Configs.valid_dataset, 'cifar10_valid_dataset'),
                       (CIFAR10Configs.train_loader, 'cifar10_train_loader'),
                       (CIFAR10Configs.valid_loader, 'cifar10_valid_loader'))



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
from torchvision import datasets, transforms

from labml import lab, tracker, experiment, logger


class Net(nn.Module):
    """
    This is the simple convolutional neural network we use for this sample
    """

    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        return self.fc2(x)


def train(model, optimizer, train_loader, device, train_log_interval):
    """This is the training code"""

    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)

        output = model(data)
        loss = F.cross_entropy(output, target)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # **✨ Increment the global step**
        tracker.add_global_step()
        # **✨ Store stats in the tracker**
        tracker.add({'loss.train': loss})

        #
        if batch_idx % train_log_interval == 0:
            # **✨ Save added stats**
            tracker.add(model=model)
            tracker.save()


def validate(model, valid_loader, device):
    """This is the validation code"""

    model.eval()

    valid_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in valid_loader:
            data, target = data.to(device), target.to(device)

            output = model(data)
            valid_loss += F.cross_entropy(output, target, reduction='sum').item()

            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()

    valid_loss /= len(valid_loader.dataset)
    valid_accuracy = 100. * correct / len(valid_loader.dataset)

    # **Save stats**
    tracker.save({'loss.valid': valid_loss, 'accuracy.valid': valid_accuracy})


def main():
    # ✨ Set the types of the stats/indicators.
    # They default to scalars if not specified
    tracker.set_queue('loss.train', 20, True)
    tracker.set_histogram('loss.valid', True)
    tracker.set_scalar('accuracy.valid', True)

    # Configurations
    configs = {
        'epochs': 10,
        'train_batch_size': 64,
        'valid_batch_size': 100,
        'use_cuda': True,
        'seed': 5,
        'train_log_interval': 10,
        'learning_rate': 0.01,
    }

    is_cuda = configs['use_cuda'] and torch.cuda.is_available()
    if not is_cuda:
        device = torch.device("cpu")
    else:
        device = torch.device(f"cuda:0")

    data_transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])

    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST(str(lab.get_data_path()),
                       train=True,
                       download=True,
                       transform=data_transform),
        batch_size=configs['train_batch_size'], shuffle=True)

    valid_loader = torch.utils.data.DataLoader(
        datasets.MNIST(str(lab.get_data_path()),
                       train=False,
                       download=True,
                       transform=data_transform),
        batch_size=configs['valid_batch_size'], shuffle=False)

    model = Net().to(device)
    optimizer = optim.Adam(model.parameters(), lr=configs['learning_rate'])

    torch.manual_seed(configs['seed'])

    # ✨ Create the experiment
    experiment.create(name='mnist_labml_tracker')

    # ✨ Save configurations
    experiment.configs(configs)

    # ✨ Set PyTorch models for checkpoint saving and loading
    experiment.add_pytorch_models(dict(model=model))

    # ✨ Start and monitor the experiment
    with experiment.start():
        #
        for epoch in range(1, configs['epochs'] + 1):
            train(model, optimizer, train_loader, device, configs['train_log_interval'])
            validate(model, valid_loader, device)
            logger.log()

    # ✨ Save the models
    experiment.save_checkpoint()


#
if __name__ == '__main__':
    main()



import importlib
import multiprocessing as mp
import multiprocessing.connection
from pathlib import PurePath
from typing import Dict, Optional

from labml import lab
from labml.internal import experiment as _experiment


class Process:
    _child: mp.connection.Connection
    _process: mp.Process

    def __init__(self, python_file: PurePath, *,
                 configs: Optional[Dict[str, any]],
                 comment: Optional[str],
                 lab_path: PurePath):
        self._child, parent = mp.Pipe()
        self._process = mp.Process(target=_run_process,
                                   args=(parent, python_file,
                                         configs, comment,
                                         lab_path))

    def __call__(self):
        self._process.start()
        ready = self._child.recv()
        assert ready == 'ready'
        print(self._process.pid)

        self._child.send("start")


def _load_module_main(python_file: PurePath):
    module_path = python_file.relative_to(lab.get_path())
    module_path = str(module_path).replace('/', '.').replace('.py', '')
    experiment_module = importlib.import_module(module_path)
    main_func = getattr(experiment_module, 'main', None)
    if main_func is None:
        raise ValueError('The experiment should have a function called main, that will be executed.')

    return main_func


def _set_arguments(*,
                   configs: Optional[Dict[str, any]],
                   comment: Optional[str]):
    _experiment.global_params_singleton().configs = configs
    _experiment.global_params_singleton().comment = comment


def new_run(python_file: PurePath, *,
            configs: Optional[Dict[str, any]],
            comment: Optional[str]):
    main_func = _load_module_main(python_file)
    _set_arguments(configs=configs, comment=comment)
    main_func()


def _run_process(remote: multiprocessing.connection.Connection,
                 python_file: PurePath,
                 configs: Optional[Dict[str, any]],
                 comment: Optional[str]):
    main_func = _load_module_main(python_file)
    _set_arguments(configs=configs, comment=comment)
    remote.send("ready")
    # sys.stdout = None

    while True:
        try:
            cmd = remote.recv()
            if cmd == "start":
                main_func()
            else:
                raise NotImplementedError
        except KeyboardInterrupt as e:
            print("Interrupted")
            print(e)
            break


def new_run_process(python_file: PurePath, *,
                    configs: Optional[Dict[str, any]],
                    comment: Optional[str]):
    p = Process(python_file,
                configs=configs,
                comment=comment,
                lab_path=lab.get_path())
    p()



import os
import pathlib
from typing import Set, List


def get_caller_file(ignore_callers: Set[str] = None):
    if ignore_callers is None:
        ignore_callers = {}

    import inspect

    frames: List[inspect.FrameInfo] = inspect.stack()
    lab_src = pathlib.PurePath(__file__).parent.parent

    for f in frames:
        module_path = pathlib.PurePath(f.filename)
        if str(module_path).startswith(str(lab_src) + '/'):
            continue
        if str(module_path) in ignore_callers:
            continue
        if str(module_path).startswith('<ipython'):
            break
        return str(module_path)

    return os.path.abspath('')


from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple, NamedTuple, Dict, Union

from labml import lab
from labml.internal import util
from labml.internal.experiment.experiment_run import RunInfo
from labml.internal.util.strings import is_pattern_match


class RunsSet:
    _runs: Dict[str, Tuple[Path, str]]

    def __init__(self):
        experiment_path = Path(lab.get_experiments_path())
        runs = {}
        for exp_path in experiment_path.iterdir():
            for run_path in exp_path.iterdir():
                runs[run_path.name] = (run_path, experiment_path.name)

        self._runs = runs

    def get(self, uuid: str) -> Tuple[RunInfo, str]:
        run_path = self._runs[uuid][0]
        run_info_path = run_path / 'run.yaml'

        with open(str(run_info_path), 'r') as f:
            data = util.yaml_load(f.read())
            run = RunInfo.from_dict(run_path.parent, data)

        return run, self._runs[uuid][1]


class IndicatorClass(Enum):
    scalar = 'scalar'
    histogram = 'histogram'
    queue = 'queue'
    tensor = 'tensor'


class StepSelect(NamedTuple):
    start: Optional[int]
    end: Optional[int]


class Indicator:
    def __init__(self, key: str, class_: IndicatorClass, uuid: str,
                 props: Dict[str, any], *,
                 select: Optional[StepSelect] = None,
                 is_mean: bool = False):
        self.uuid = uuid
        self.class_ = class_
        self.key = key
        self.props = props
        if class_ == IndicatorClass.tensor and props.get('is_once', False):
            select = None
        if select is None:
            select = StepSelect(None, None)
        self.select = select
        self.is_mean = is_mean

    def hash_str(self):
        return f"{self.uuid}#{self.key}"

    @property
    def is_distribution(self):
        return self.class_ in [IndicatorClass.histogram, IndicatorClass.queue]

    @property
    def is_scalar(self):
        return self.class_ == IndicatorClass.scalar


class IndicatorCollection:
    _indicators: List[Indicator]

    def __init__(self, indicators: List[Indicator]):
        has = set()
        self._indicators = []
        for ind in indicators:
            h = ind.hash_str()
            if h in has:
                continue
            has.add(h)
            self._indicators.append(ind)

        self._indicator_keys = {ind.key.replace('.', '_'): ind.key for ind in self._indicators}
        self._indicators_list = [k for k in self._indicator_keys.keys()]

    def __dir__(self):
        return self._indicators_list

    def __getattr__(self, k: str):
        key = self._indicator_keys[k]
        inds = []
        for ind in self._indicators:
            if ind.key == key:
                inds.append(ind)

        return IndicatorCollection(inds)

    def __add__(self, other: 'IndicatorCollection'):
        return IndicatorCollection(self._indicators + other._indicators)

    def __radd__(self, other: Optional['IndicatorCollection']):
        if other is None:
            return IndicatorCollection(self._indicators)
        else:
            return IndicatorCollection(self._indicators + other._indicators)

    def __iter__(self):
        return iter(self._indicators)

    def __len__(self):
        return len(self._indicators)

    def __getitem__(self, item: Union[slice, str]):
        if isinstance(item, slice):
            select = StepSelect(item.start, item.stop)
            inds = [Indicator(ind.key, ind.class_, ind.uuid, ind.props,
                              select=select, is_mean=ind.is_mean)
                    for ind in self._indicators]
            return IndicatorCollection(inds)
        elif isinstance(item, str):
            inds = [ind for ind in self._indicators if is_pattern_match(ind.key, item)]
            return IndicatorCollection(inds)
        else:
            raise ValueError(item)

    def mean(self):
        inds = [Indicator(ind.key, ind.class_, ind.uuid, ind.props,
                          select=ind.select,
                          is_mean=True) for ind in self._indicators]
        return IndicatorCollection(inds)


class Run:
    indicators: IndicatorCollection
    name: str
    run_info: RunInfo

    def __init__(self, uuid: str):
        runs = RunsSet()
        self.run_info, self.name = runs.get(uuid)

        with open(str(self.run_info.indicators_path), 'r') as f:
            indicators = util.yaml_load(f.read())

        if 'indicators' not in indicators:
            raise RuntimeError("This run is corrupted or from an old version of LabML. "
                               "Please update labml_dashboard and run it on this project. "
                               "It will automatically migrate all the experiments.")

        indicators = indicators['indicators']

        inds = []
        for k, v in indicators.items():
            cn = v['class_name']
            class_ = None
            if cn == 'Histogram':
                class_ = IndicatorClass.histogram
            elif cn == 'Queue':
                class_ = IndicatorClass.queue
            elif cn == 'IndexedScalar':
                class_ = IndicatorClass.scalar
            elif cn == 'Scalar':
                class_ = IndicatorClass.scalar
            elif cn == 'Tensor':
                class_ = IndicatorClass.tensor

            if class_ is None:
                continue
            inds.append(Indicator(k, class_, self.run_info.uuid, v))

        self.indicators = IndicatorCollection(inds)



from typing import Dict

from ..indicators import Indicator


class Writer:
    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):
        raise NotImplementedError()

    def flush(self):
        pass

    def write_h_parameters(self, hparams: Dict[str, any]):
        pass

    def save_indicators(self, dot_indicators: Dict[str, Indicator], indicators: Dict[str, Indicator]):
        pass



from pathlib import Path
from typing import Dict, Union, Optional, List

from labml.internal import util
from .base import Configs
from .utils import Value

_CONFIG_PRINT_LEN = 40


class ConfigsSaver:
    def save(self, configs: Dict):
        raise NotImplementedError


class FileConfigsSaver(ConfigsSaver):
    def __init__(self, configs_path: Path):
        self.configs_path = configs_path

    def save(self, configs: Dict):
        with open(str(self.configs_path), "w") as file:
            file.write(util.yaml_dump(configs))


class ConfigProcessor:
    configs_dict: Optional[Dict[str, any]]
    configs: Optional[Configs]
    savers: List[ConfigsSaver]

    def __init__(self, configs: Union[Configs, Dict[str, any]], values: Dict[str, any] = None):
        if values is None:
            values = {}

        if isinstance(configs, Configs):
            self.configs_dict = None
            self.configs = configs
            self.configs._set_update_callback(self._on_configs_updated)
            self.configs._set_values(values)
        elif isinstance(configs, dict):
            self.configs = None
            self.configs_dict = configs
            self.configs_dict.update(values)

        self.values = values
        self.savers = []

    def _on_configs_updated(self):
        configs = self.to_json()
        for s in self.savers:
            s.save(configs)

    def to_json(self):
        if self.configs is not None:
            return self.configs._to_json()
        else:
            orders = {k: i for i, k in enumerate(self.configs_dict.keys())}
            configs = {}
            for k, v in self.configs_dict.items():
                configs[k] = {
                    'name': k,
                    'type': str(type(v)),
                    'value': Value.to_yaml_truncated(v),
                    'order': orders.get(k, -1),
                    'options': [],
                    'computed': Value.to_yaml_truncated(v),
                    'is_hyperparam': False,
                    'is_meta': False,
                    'is_explicitly_specified': (k in self.values)
                }

            return configs

    def add_saver(self, saver: ConfigsSaver):
        self.savers.append(saver)
        saver.save(self.to_json())

    def get_hyperparams(self):
        hyperparams = {}
        for k, v in self.to_json().items():
            if v['is_hyperparam'] or v['is_explicitly_specified']:
                value = v['value'] or v['computed']
                if type(value) not in {int, float, str}:
                    value = Value.to_str(value)

                hyperparams[k] = value

        return hyperparams


def load_configs(configs_path: Path):
    if not configs_path.exists():
        return None

    with open(str(configs_path), 'r') as file:
        configs = util.yaml_load(file.read())

    return configs



"""
---
title: Train Autoregressive Transformer
summary: This is training code with notes for a basic auto-regressive transformer.
---

# Train Autoregressive Transformer

This trains a simple [transformer](../../) model for auto-regression.
"""

from typing import Callable, Any

import torch
import torch.nn as nn
from torchtext.data.utils import get_tokenizer

from labml import lab, experiment, monit, tracker, logger
from labml.configs import option
from labml.logger import Text
from labml.utils.pytorch import get_modules
from labml_helpers.datasets.text import TextDataset, SequentialDataLoader, TextFileDataset
from labml_helpers.metrics.accuracy import Accuracy
from labml_helpers.module import Module
from labml_helpers.optimizer import OptimizerConfigs
from labml_helpers.train_valid import SimpleTrainValidConfigs, BatchIndex
from labml_nn.transformers import Encoder, Generator, TransformerConfigs
from labml_nn.transformers.utils import subsequent_mask


class AutoregressiveModel(Module):
    """
    ## Auto regressive model
    """

    def __init__(self, src_embed: Module, encoder: Encoder, generator: Generator, *,
                 is_save_ff_input: bool = False):
        super().__init__()
        # Token embedding module
        self.src_embed = src_embed
        # Transformer based encoder
        self.encoder = encoder
        # Whether the last layer of the encoder should
        # save the input to the feed-forward layer.
        # This is out $f(c_t)$, the embedding of the context.
        self.encoder.layers[-1].is_save_ff_input = is_save_ff_input
        # Next token generation layer;
        # this give logits  of the the next token
        self.generator = generator
        # This will be initialized on the first call
        self.src_mask = None

    @property
    def ff_input(self) -> torch.Tensor:
        """
        Retrieve saved $f(c_t)$
        """
        return self.encoder.layers[-1].ff_input

    def __call__(self, src: torch.Tensor):
        # Create subsequent mask, so that the transformer can only pay attention to past tokens.
        if self.src_mask is None or self.src_mask.size(0) != len(src):
            self.src_mask = subsequent_mask(len(src)).to(src.device)
        # Embed the tokens (`src`) and run it through the the transformer
        res = self.encoder(self.src_embed(src), self.src_mask)
        # Generate logits of the next token
        return self.generator(res)


class CrossEntropyLoss(Module):
    """
    Cross entropy loss
    """

    def __init__(self):
        super().__init__()
        self.loss = nn.CrossEntropyLoss()

    def __call__(self, outputs, targets):
        return self.loss(outputs.view(-1, outputs.shape[-1]), targets.view(-1))


class Configs(SimpleTrainValidConfigs):
    """
    ## Configurations

    The default configs can and will be over-ridden when we start the experiment
    """

    transformer: TransformerConfigs
    model: AutoregressiveModel
    text: TextDataset
    batch_size: int = 20
    seq_len: int = 32
    n_tokens: int
    tokenizer: Callable = 'character'

    is_save_models = True
    prompt: str
    prompt_separator: str

    is_save_ff_input = False
    optimizer: torch.optim.Adam = 'transformer_optimizer'

    accuracy = Accuracy()
    loss_func = CrossEntropyLoss()

    def init(self):
        # Create a configurable optimizer.
        # Parameters like learning rate can be changed by passing a dictionary when starting the experiment.
        optimizer = OptimizerConfigs()
        optimizer.parameters = self.model.parameters()
        optimizer.d_model = self.transformer.d_model
        optimizer.optimizer = 'Noam'
        self.optimizer = optimizer

        # Create a sequential data loader for training
        self.train_loader = SequentialDataLoader(text=self.text.train,
                                                 dataset=self.text,
                                                 batch_size=self.batch_size,
                                                 seq_len=self.seq_len)

        # Create a sequential data loader for validation
        self.valid_loader = SequentialDataLoader(text=self.text.valid,
                                                 dataset=self.text,
                                                 batch_size=self.batch_size,
                                                 seq_len=self.seq_len)

        self.state_modules = [self.accuracy]

    def sample(self):
        """
        Sampling function to generate samples periodically while training
        """
        prompt = self.prompt
        log = [(prompt, Text.subtle)]
        # Sample 25 tokens
        for i in monit.iterate('Sample', 25):
            # Tokenize the prompt
            data = self.text.text_to_i(prompt).unsqueeze(-1)
            data = data.to(self.device)
            # Get the model output
            output = self.model(data)
            # Get the model prediction (greedy)
            output = output.argmax(dim=-1).squeeze()
            # Add the prediction to prompt
            prompt += self.prompt_separator + self.text.itos[output[-1]]
            # Add the prediction for logging
            log += [(self.prompt_separator + self.text.itos[output[-1]], Text.value)]

        logger.log(log)

    def step(self, batch: Any, batch_idx: BatchIndex):
        """
        This method is called for each batch
        """
        self.model.train(self.mode.is_train)

        # Get data and target labels
        data, target = batch[0].to(self.model.device), batch[1].to(self.model.device)

        if self.mode.is_train:
            tracker.add_global_step(data.shape[0] * data.shape[1])

        # Run the model
        output = self.model(data)

        # Calculate loss
        loss = self.loss_func(output, target)
        # Calculate accuracy
        self.accuracy(output, target)

        # Log the loss
        tracker.add("loss.", loss)

        #  If we are in training mode, calculate the gradients
        if self.mode.is_train:
            loss.backward()
            self.optimizer.step()
            if batch_idx.is_last:
                tracker.add('model', self.model)
            self.optimizer.zero_grad()

        tracker.save()


@option(Configs.tokenizer)
def basic_english():
    """
    Basic  english tokenizer

    We use character level tokenizer in this experiment.
    You can switch by setting,

    ```
        'tokenizer': 'basic_english',
    ```

    as the configurations dictionary when starting the experiment.

    """
    return get_tokenizer('basic_english')


def character_tokenizer(x: str):
    return list(x)


@option(Configs.tokenizer)
def character():
    """
    Character level tokenizer
    """
    return character_tokenizer


@option(Configs.text)
def tiny_shakespeare(c: Configs):
    """
    Initialize/load tiny shakespeare dataset

    This dataset is from Andrej Karpathy's [char-rnn](https://github.com/karpathy/char-rnn) project.
    """
    return TextFileDataset(
        lab.get_data_path() / 'tiny_shakespeare.txt', c.tokenizer,
        url='https://raw.githubusercontent.com/karpathy/char-rnn/master/data/tinyshakespeare/input.txt')


@option(Configs.model)
def autoregressive_model(c: Configs):
    """
    Initialize the auto-regressive model
    """
    m = AutoregressiveModel(
        # Get the source token embedding layer, encoder and
        # final token generator from configurable transformer
        src_embed=c.transformer.src_embed,
        encoder=c.transformer.encoder,
        generator=c.transformer.generator,
        # Whether to save $f(c_t)$
        is_save_ff_input=c.is_save_ff_input)
    return m.to(c.device)


@option(Configs.transformer)
def transformer_c(c: Configs):
    """
    Initialize the configurable transformer encoder for our autoregressive model
    """
    tc = TransformerConfigs()
    tc.n_src_vocab = c.n_tokens
    tc.n_tgt_vocab = c.n_tokens

    return tc


def main():
    # Create experiment
    experiment.create(name="knn_lm", comment='', writers={'tensorboard', 'sqlite', 'screen'})
    # Create configs
    conf = Configs()
    # Load configurations
    experiment.configs(conf,
                       # A dictionary of configurations to override
                       {'tokenizer': 'character',
                        'prompt_separator': '',
                        'prompt': 'It is ',
                        'text': 'tiny_shakespeare',

                        'seq_len': 1024,
                        'epochs': 128,
                        'batch_size': 6,
                        'inner_iterations': 10,

                        # Transformer configurations
                        'transformer.d_model': 256,
                        'transformer.d_ff': 1024,
                        'transformer.n_heads': 8,
                        'transformer.n_layers': 6})

    # This is needed to initialize models
    conf.n_tokens = conf.text.n_tokens

    # Set models for saving and loading
    experiment.add_pytorch_models(get_modules(conf))

    # Start the experiment
    with experiment.start():
        # `TrainValidConfigs.run`
        conf.run()


if __name__ == '__main__':
    main()



from pathlib import PurePath
from typing import Dict

import tensorflow as tf
from tensorboard.plugins.hparams import api as hp

from . import TensorboardWriter

tf.config.experimental.set_visible_devices([], "GPU")


class TensorflowTensorboardWriter(TensorboardWriter):
    def __init__(self, log_path: PurePath):
        super().__init__()

        self.__log_path = log_path
        self.__writer = None

    def __connect(self):
        if self.__writer is not None:
            return

        self.__writer = tf.summary.create_file_writer(str(self.__log_path))

    def hparams(self, hparams: Dict[str, any]):
        self.__connect()
        with self.__writer.as_default():
            hp.hparams(hparams)

    def add_scalar(self, key: str, value: any, global_step: int):
        self.__connect()
        with self.__writer.as_default():
            tf.summary.scalar(key, value, step=global_step)

    def add_histogram(self, key: str, value: any, global_step: int):
        self.__connect()
        with self.__writer.as_default():
            tf.summary.histogram(key, value, step=global_step)

    def add_image(self, key: str, img: any, global_step: int):
        self.__connect()
        img = img.cpu().detach().permute(0, 2, 3, 1).numpy()
        with self.__writer.as_default():
            tf.summary.image(key, img, step=global_step)



from pathlib import PurePath, Path
from typing import List, Optional, Dict

from labml.internal import util
from labml.internal.util import is_colab, is_kaggle
from labml.logger import Text
from labml.utils import get_caller_file
from labml.utils.notice import labml_notice

_CONFIG_FILE_NAME = '.labml.yaml'


class LabYamlNotfoundError(RuntimeError):
    pass


class WebAPIConfigs:
    url: str
    frequency: float
    verify_connection: bool
    open_browser: bool

    def __init__(self, *,
                 url: str,
                 frequency: float,
                 verify_connection: bool,
                 open_browser: bool):
        self.open_browser = open_browser
        self.frequency = frequency
        self.verify_connection = verify_connection
        self.url = url


class Lab:
    """
    ### Lab

    Lab contains the labml specific properties.
    """
    experiments: Optional[Path]
    data_path: Optional[Path]
    check_repo_dirty: Optional[bool]
    path: Optional[Path]
    web_api: Optional[WebAPIConfigs]

    def __init__(self):
        self.indicators = {}
        self.path = None
        self.check_repo_dirty = None
        self.data_path = None
        self.experiments = None
        self.web_api = None
        self.configs = self.__default_config()
        self.custom_configs = []
        self.__update_configs()
        self.__current_path = None

        python_file = get_caller_file()
        self.__load_configs(Path(python_file).resolve())

    def set_path(self, path: str):
        self.__load_configs(Path(path).resolve())

    def __load_configs(self, path: Path):
        if path == self.__current_path:
            return
        self.__current_path = path
        config_files = self.__load_config_files(path)

        if not config_files:
            if not is_colab() and not is_kaggle():
                labml_notice([(".labml.yaml", Text.value),
                              " config file could not be found. Looking in path: ",
                              (path, Text.meta)])
                while path.exists() and not path.is_dir():
                    path = path.parent

        for c in config_files:
            self.__merge_configs(c)

        for c in self.custom_configs:
            self.__merge_configs(c)

        if not config_files and self.configs['path'] is None:
            self.configs['path'] = str(path)

        self.__update_configs()

    def __update_configs(self):
        if self.configs['path'] is None:
            self.path = None
            self.experiments = None
            self.data_path = None
        else:
            self.path = Path(self.configs['path'])
            self.data_path = (self.path / self.configs['data_path']).resolve()
            self.experiments = (self.path / self.configs['experiments_path']).resolve()

        self.check_repo_dirty = self.configs['check_repo_dirty']
        self.indicators = self.configs['indicators']
        if self.configs['web_api']:
            web_api_url = self.configs['web_api']
            if web_api_url[0:4] != 'http':
                web_api_url = f"https://api.lab-ml.com/api/v1/track?labml_token={web_api_url}&"
            self.web_api = WebAPIConfigs(url=web_api_url,
                                         frequency=self.configs['web_api_frequency'],
                                         verify_connection=self.configs['web_api_verify_connection'],
                                         open_browser=self.configs['web_api_open_browser'])
        else:
            self.web_api = None

    def set_configurations(self, configs: Dict[str, any]):
        self.custom_configs.append(configs)
        for c in self.custom_configs:
            self.__merge_configs(c)
        self.__update_configs()

    def __str__(self):
        return f"<Lab path={self.path}>"

    def __repr__(self):
        return str(self)

    @staticmethod
    def __default_config():
        return dict(
            path=None,
            check_repo_dirty=False,
            config_file_path=None,
            data_path='data',
            experiments_path='logs',
            analytics_path='analytics',
            analytics_templates={},
            web_api='https://api.lab-ml.com/api/v1/track?',
            web_api_frequency=60,
            web_api_verify_connection=True,
            web_api_open_browser=True,
            indicators=[{
                'class_name': 'Scalar',
                'is_print': True,
                'name': '*'
            }, {
                'class_name': 'Scalar',
                'is_print': False,
                'name': 'param.*'
            }, {
                'class_name': 'Scalar',
                'is_print': False,
                'name': 'grad.*'
            }, {
                'class_name': 'Scalar',
                'is_print': False,
                'name': 'module.*'
            }, {
                'class_name': 'Scalar',
                'is_print': False,
                'name': 'optim.*'
            }, {
                'class_name': 'Scalar',
                'is_print': False,
                'name': 'time.*'
            }
            ]
        )

    def __merge_configs(self, c):
        if self.configs['path'] is None and 'config_file_path' in c:
            self.configs['path'] = c['config_file_path']

        for k, v in c.items():
            if k not in self.configs:
                raise RuntimeError(f"Unknown config parameter #{k} in file "
                                   f"{c['config_file_path'] / _CONFIG_FILE_NAME}")
            elif k == 'indicators':
                self.configs[k] += v
            else:
                self.configs[k] = v

    @staticmethod
    def __load_config_files(path: Path):
        configs = []

        while path.exists():
            if path.is_dir():
                config_file = path / _CONFIG_FILE_NAME
                if config_file.is_file():
                    with open(str(config_file)) as f:
                        config = util.yaml_load(f.read())
                        if config is None:
                            config = {}
                        config['config_file_path'] = path
                        configs.append(config)

            if str(path) == path.root:
                break

            path = path.parent

        return configs

    def get_experiments(self) -> List[Path]:
        """
        Get list of experiments
        """
        experiments_path = Path(self.experiments)
        return [child for child in experiments_path.iterdir()]


_internal: Optional[Lab] = None


def lab_singleton() -> Lab:
    global _internal
    if _internal is None:
        _internal = Lab()

    return _internal



import json
import threading
import time
from pathlib import PurePath
from queue import Queue
from typing import Dict, Optional

import numpy as np

from . import Writer as WriteBase
from ..indicators import Indicator
from ..indicators.numeric import NumericIndicator

MAX_BUFFER_SIZE = 1024
WARMUP_COMMITS = 5
TIMEOUT_SECONDS = 5
FREQUENCY = 60


class FileWriterThread(threading.Thread):
    def __init__(self, file_path: PurePath):
        super().__init__(daemon=False)
        self.file_path = file_path
        self.queue = Queue()

    def push(self, data: any):
        self.queue.put(data)

    def run(self):
        while True:
            data = self.queue.get()
            if data == 'done':
                return
            else:
                self._process(data)

    def _process(self, data: Dict[str, any]):
        res = self._send(data['packet'])
        if 'accept' in data:
            data['accept'](res)

    def _send(self, data: Dict[str, any]):
        with open(str(self.file_path), 'a') as f:
            f.write(json.dumps(data) + '\n')


class Writer(WriteBase):
    url: Optional[str]
    thread: Optional[FileWriterThread]

    def __init__(self, file_path: PurePath):
        super().__init__()

        self.file_path = file_path
        self.scalars_cache = []
        self.last_committed = time.time()
        self.commits_count = 0
        self.indicators = {}

        self.thread = FileWriterThread(file_path)

    @staticmethod
    def _parse_key(key: str):
        return key

    def _write_indicator(self, global_step: int, indicator: Indicator):
        if indicator.is_empty():
            return

        if isinstance(indicator, NumericIndicator):
            mean_value = indicator.get_mean()
            hist = indicator.get_histogram()
            if hist:
                if len(hist) > 1:
                    hist, bins = np.histogram(hist)
                    hist = {'hist': hist.tolist(), 'bins': bins.tolist()}
                else:
                    hist = None
            key = self._parse_key(indicator.mean_key)
            if key not in self.indicators:
                self.indicators[key] = {
                    'mean': [],
                    'hist': []
                }

            self.indicators[key]['mean'].append((global_step, mean_value))
            self.indicators[key]['hist'].append(hist)

    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):
        for ind in indicators.values():
            # if not ind.is_print:
            #     continue
            self._write_indicator(global_step, ind)

        t = time.time()
        freq = FREQUENCY
        if self.commits_count < WARMUP_COMMITS:
            freq /= 2 ** (WARMUP_COMMITS - self.commits_count)
        if t - self.last_committed > freq:
            self.last_committed = t
            self.commits_count += 1
            self.flush()

    def flush(self):
        data = {}

        for key, ind_values in self.indicators.items():
            value = np.array(ind_values['mean'])
            step: np.ndarray = value[:, 0]
            value: np.ndarray = value[:, 1]
            while value.shape[0] > MAX_BUFFER_SIZE:
                if value.shape[0] % 2 == 1:
                    value = np.concatenate((value, value[-1:]))
                    step = np.concatenate((step, step[-1:]))

                n = value.shape[0] // 2
                step = np.mean(step.reshape(n, 2), axis=-1)
                value = np.mean(value.reshape(n, 2), axis=-1)

            data[key] = {
                'step': step.tolist(),
                'value': value.tolist(),
                'hist': ind_values['hist']
            }

        self.indicators = {}

        self.thread.push({'packet': {
            'track': data,
            'time': time.time()
        }})
        if not self.thread.is_alive():
            self.thread.start()



from pathlib import PurePath
from typing import Dict, List, Optional, Callable, Union, Tuple

from labml.internal import util
from labml.internal.lab import lab_singleton, LabYamlNotfoundError
from labml.internal.util import strings
from .indicators import Indicator
from .indicators.factory import load_indicator_from_dict, create_default_indicator
from .indicators.numeric import Scalar
from .namespace import Namespace
from .writers import Writer
from .writers.screen import ScreenWriter
from ..logger import LogPart
from labml.internal.util.colors import StyleCode
from ... import logger
from ...logger import Text


class Tracker:
    __loop_counter: int
    __set_looping_indicators: Optional[Callable[[List[Union[str, Tuple[str, Optional[StyleCode]]]]], None]]

    dot_indicators: Dict[str, Indicator]
    namespaces: List[Namespace]
    indicators: Dict[str, Indicator]

    def __init__(self):
        self.__store = None
        self.__writers: List[Writer] = []

        self.__start_global_step: Optional[int] = None
        self.__global_step: Optional[int] = None
        self.__last_global_step: Optional[int] = None

        self.__set_looping_indicators = None
        self.__loop_counter = 0

        self.indicators = {}
        self.dot_indicators = {}
        self.__indicators_file = None
        self.namespaces = []
        self.is_indicators_updated = True
        self.reset_store()

        from labml.internal.tracker.writers import screen
        self.add_writer(screen.ScreenWriter())

    def __assert_name(self, name: str, value: any):
        if name.endswith("."):
            if name in self.dot_indicators:
                assert self.dot_indicators[name].equals(value)

        assert name not in self.indicators, f"{name} already used"

    def add_writer(self, writer: Writer):
        self.__writers.append(writer)

    def reset_writers(self):
        self.__writers = []

    def write_h_parameters(self, hparams: Dict[str, any]):
        for w in self.__writers:
            w.write_h_parameters(hparams)

    def _write_writer(self, writer: Writer, global_step):
        return writer.write(global_step=global_step,
                            indicators=self.indicators)

    def clear(self):
        for k, v in self.indicators.items():
            v.clear()

    def write(self):
        global_step = self.global_step

        self.save_indicators()

        indicators_print = None

        for w in self.__writers:
            if isinstance(w, ScreenWriter):
                indicators_print = self._write_writer(w, global_step)
            else:
                self._write_writer(w, global_step)
        self.clear()

        if indicators_print is not None:
            if self.__is_looping:
                self.__set_looping_indicators(indicators_print)
            else:
                parts = [(f"{self.global_step :8,}:  ", Text.highlight)]
                parts += indicators_print
                logger.log(parts, is_new_line=False)

    @property
    def global_step(self) -> int:
        if self.__global_step is not None:
            return self.__global_step

        global_step = 0
        if self.__start_global_step is not None:
            global_step = self.__start_global_step

        if self.__is_looping:
            return global_step + self.__loop_counter

        if self.__last_global_step is not None:
            return self.__last_global_step

        return global_step

    def reset_store(self):
        self.indicators = {}
        self.dot_indicators = {}
        self.__indicators_file = None
        self.namespaces = []
        self.is_indicators_updated = True
        try:
            for ind in lab_singleton().indicators:
                self.add_indicator(load_indicator_from_dict(ind))
        except LabYamlNotfoundError:
            pass

    def add_indicator(self, indicator: Indicator):
        self.dot_indicators[indicator.name] = indicator
        self.is_indicators_updated = True

    def save_indicators(self, file: Optional[PurePath] = None):
        if not self.is_indicators_updated:
            return
        self.is_indicators_updated = False
        if file is None:
            if self.__indicators_file is None:
                return
            file = self.__indicators_file
        else:
            self.__indicators_file = file

        wildcards = {k: ind.to_dict() for k, ind in self.dot_indicators.items()}
        inds = {k: ind.to_dict() for k, ind in self.indicators.items()}
        with open(str(file), "w") as file:
            file.write(util.yaml_dump({'wildcards': wildcards,
                                       'indicators': inds}))

        for w in self.__writers:
            w.save_indicators(self.dot_indicators, self.indicators)

    def _create_indicator(self, key: str, value: any):
        if key in self.indicators:
            return

        ind_key, ind_score = strings.find_best_pattern(key, self.dot_indicators.keys())
        if ind_key is None:
            raise ValueError(f"Cannot find matching indicator for {key}")
        if ind_score == 0:
            is_print = self.dot_indicators[ind_key].is_print
            self.indicators[key] = create_default_indicator(key, value, is_print)
        else:
            self.indicators[key] = self.dot_indicators[ind_key].copy(key)
        self.is_indicators_updated = True

    def store(self, key: str, value: any):
        if key.endswith('.'):
            key = '.'.join([key[:-1]] + [ns.name for ns in self.namespaces])

        self._create_indicator(key, value)
        self.indicators[key].collect_value(value)

    def new_line(self):
        for w in self.__writers:
            if isinstance(w, ScreenWriter):
                logger.log()

    def namespace(self, name: str):
        return Namespace(tracker=self, name=name)

    def namespace_enter(self, ns: Namespace):
        self.namespaces.append(ns)

    def namespace_exit(self, ns: Namespace):
        if len(self.namespaces) == 0:
            raise RuntimeError("Impossible")

        if ns is not self.namespaces[-1]:
            raise RuntimeError("Impossible")

        self.namespaces.pop(-1)

    def set_global_step(self, global_step: Optional[int]):
        self.__global_step = global_step

    def set_start_global_step(self, global_step: Optional[int]):
        self.__start_global_step = global_step

    def add_global_step(self, increment_global_step: int = 1):
        if self.__global_step is None:
            self.__global_step = self.global_step

        self.__global_step += increment_global_step

    @property
    def __is_looping(self):
        return self.__set_looping_indicators is not None

    def start_loop(self, set_looping_indicators: Callable[[List[LogPart]], None]):
        self.__set_looping_indicators = set_looping_indicators

    def loop_count(self, value: int):
        self.__loop_counter = value

    def finish_loop(self):
        self.__last_global_step = self.global_step
        self.__set_looping_indicators = None
        for w in self.__writers:
            w.flush()


_internal: Optional[Tracker] = None


def tracker_singleton() -> Tracker:
    global _internal
    if _internal is None:
        _internal = Tracker()

    return _internal



from typing import List, Tuple

import torch
from labml_helpers.module import Module
from torch import nn

from labml_samples.pytorch.stocks import CandleIdx


class FrontPaddedConv1d(Module):
    def __init__(self, *, in_channels: int, out_channels: int, kernel_size: int):
        super().__init__()

        self.kernel_size = kernel_size
        self.conv = nn.Conv1d(in_channels=in_channels,
                              out_channels=out_channels,
                              kernel_size=kernel_size)

    def __call__(self, x: torch.Tensor):
        if self.kernel_size > 1:
            pad = x.new_zeros(*x.shape[:-1], self.kernel_size - 1)
            x = torch.cat([pad, x], dim=-1)
        return self.conv(x)


class CnnModel(Module):
    def __init__(self, *,
                 price_mean: float,
                 price_std: float,
                 volume_mean: float,
                 volume_std: float,
                 y_std: float,
                 y_mean: float,
                 activation: nn.Module,
                 conv_sizes: List[Tuple[int, int]],
                 dropout=0.1):
        super().__init__()

        self.y_mean = y_mean
        self.y_std = y_std
        self.volume_std = volume_std
        self.volume_mean = volume_mean
        self.price_std = price_std
        self.price_mean = price_mean
        layers = []
        in_channels = CandleIdx.all
        for s in conv_sizes:
            layers.append(FrontPaddedConv1d(in_channels=in_channels,
                                            out_channels=s[0],
                                            kernel_size=s[1]))
            in_channels = s[0]
        self.layers = nn.ModuleList(layers)

        self.final = FrontPaddedConv1d(in_channels=in_channels,
                                       out_channels=1,
                                       kernel_size=1)
        self.activation = activation

        self.conv_dropout = nn.Dropout(dropout)

    def __call__(self, x: torch.Tensor):
        x = x.permute((0, 2, 1))
        x[:, 0:CandleIdx.prices, :] = \
            (x[:, 0:CandleIdx.prices, :] - self.price_mean) / self.price_std
        x[:, CandleIdx.volume, :] = \
            (x[:, CandleIdx.prices, :] - self.volume_mean) / self.volume_std
        for layer in self.layers:
            x = self.activation(self.conv_dropout(layer(x)))
        x = self.final(x)

        x = x.permute((0, 2, 1))
        return x



import urllib.request

import numpy as np
import pandas as pd
from labml import monit, lab, logger

from labml_samples.pytorch.stocks import CandleIdx


def parse(df: pd.DataFrame):
    time = np.zeros((len(df)), dtype=int)
    date = []
    for i in monit.iterate("Calculate time", len(df)):
        hour = int(df['Time'][i][0:2])
        mint = int(df['Time'][i][3:5])
        time[i] = hour * 60 + mint
        mon = df['Date'][i][0:2]
        day = df['Date'][i][3:5]
        year = df['Date'][i][6:10]
        date.append(f"{year}-{mon}-{day}")

    time = time - 570
    df['Minute'] = time
    df['Date'] = date

    return df


def filter_premarket(df: pd.DataFrame):
    df = df[0 <= df['Minute']]
    df = df[df['Minute'] < 390]
    df = df.reset_index()
    df = df[['Date', 'Time', 'Minute', 'Open', 'High', 'Low', 'Close', 'Volume']]

    return df


def to_daily_packets(df: pd.DataFrame):
    volume = np.array(df['Volume'], dtype=float)
    time = np.array(df["Minute"])
    candles = np.zeros((len(df), 6), dtype=float)
    candles[:, CandleIdx.high] = np.array(df['High'])
    candles[:, CandleIdx.low] = np.array(df['Low'])
    candles[:, CandleIdx.open] = np.array(df['Open'])
    candles[:, CandleIdx.close] = np.array(df['Close'])
    candles[:, CandleIdx.volume] = volume
    candles[:, 5] = time

    dates = []
    packets = []
    current_day = None
    packet = None
    for i in monit.iterate("To daily packets", len(df)):
        d = df['Date'][i]
        if d != current_day:
            if current_day is not None:
                dates.append(current_day)
                packets.append(packet)
            current_day = d
            packet = np.zeros((390, 5), dtype=float)
        t = time[i]
        if 0 <= t < 390:
            packet[t, :] = candles[i, 0:5]
    if current_day is not None:
        dates.append(current_day)
        packets.append(packet)

    return np.array(dates), np.array(packets)


def fill_empty_minutes_in_packet(packet: np.ndarray):
    last_idx = None
    for i in range(390):
        if np.sum(packet[i] == 0) > 0:
            if last_idx is not None:
                packet[i, 0:4] = packet[last_idx, 0:4]
        if np.sum(packet[i] > 0) == len(packet[i]):
            if last_idx is None:
                for j in range(i):
                    packet[j, 0:4] = packet[i, 0:4]
            last_idx = i


def fill_empty_minutes_in_packets(packets: np.ndarray):
    for i in monit.iterate("Fill empty minutes", packets.shape[0]):
        fill_empty_minutes_in_packet(packets[i])


def to_numpy(df: pd.DataFrame):
    dates, packets = to_daily_packets(df)
    empty_mins = np.sum(packets[:, :, 4] == 0)
    filled_mins = np.sum(packets[:, :, 4] > 0)
    empty_mins_high_activity = np.sum(packets[:, :, 4] == 0)
    fill_empty_minutes_in_packets(packets)
    zero_price = np.sum(packets[:, :, 0:4] == 0)
    zero_volume = np.sum(packets[:, :, 4] == 0)
    packets[:, :, 4] = np.maximum(packets[:, :, 4], 1)

    logger.inspect(empty_mins=empty_mins,
                   filled_mins=filled_mins,
                   empty_mins_high_activity=empty_mins_high_activity,
                   zero_price=zero_price,
                   zero_volume=zero_volume)

    return dates, packets


def build_cache(*,
                filename: str = 'IBM_unadjusted.txt',
                url: str = 'http://api.kibot.com/?action=history&symbol=IBM&interval=1&unadjusted=0&bp=1&user=guest',
                force_download: bool = False):
    data_path = lab.get_data_path() / filename
    data_with_header = lab.get_data_path() / 'stocks.csv'

    if not lab.get_data_path().exists():
        lab.get_data_path().mkdir(parents=True)

    if force_download or not data_path.exists():
        data_with_header.unlink(True)
        with monit.section('Download data') as s:
            def reporthook(count, block_size, total_size):
                s.progress(count * block_size / total_size)

            urllib.request.urlretrieve(url, str(data_path), reporthook=reporthook)

    if not data_with_header.exists():
        with open(str(data_with_header), 'w') as fh:
            fh.write('Date,Time,Open,High,Low,Close,Volume\n')
            with open(str(data_path), 'r') as f:
                fh.write(f.read())

    with monit.section("Read data"):
        df = pd.read_csv(str(data_with_header))
    df = parse(df)
    with monit.section("Filter pre-market data"):
        df = filter_premarket(df)

    with monit.section("To numpy"):
        dates, packets = to_numpy(df)

    with monit.section("Save"):
        np.save(str(lab.get_data_path() / "packets.npy"), packets)
        np.save(str(lab.get_data_path() / "dates.npy"), dates)


if __name__ == '__main__':
    build_cache()



import warnings
from typing import Tuple, Union, List, TYPE_CHECKING

if TYPE_CHECKING:
    from labml.internal.configs.config_item import ConfigItem

TRUNCATED_TOKEN = '[[...]]'
MIN_STR_LIMIT = 5


class Value:
    @staticmethod
    def is_primitive(value):
        if value is None:
            return True

        if (isinstance(value, str)
                or isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, bool)
        ):
            return True

        return False

    @staticmethod
    def __to_yaml_truncated_list(value, limit: int):
        res = []
        total_size = 0
        for v in value:
            r, size = Value.__to_yaml_truncated(v, limit - total_size)
            res.append(r)
            if limit - total_size < size:
                res.append(TRUNCATED_TOKEN)
                return res, limit
            total_size += size
        return res, total_size

    @staticmethod
    def __to_yaml_truncated_str(value, limit: int):
        limit = max(limit, MIN_STR_LIMIT)
        if len(value) > limit:
            res = value[:limit] + ' ' + TRUNCATED_TOKEN
            return res, limit
        else:
            return value, max(len(value), 1)

    @staticmethod
    def __to_yaml_truncated(value, limit: int) -> Tuple[any, int]:
        if isinstance(value, str):
            return Value.__to_yaml_truncated_str(value, limit)
        elif Value.is_primitive(value):
            return value, 1
        elif isinstance(value, list):
            return Value.__to_yaml_truncated_list(value, limit)
        elif isinstance(value, dict):
            res = {}
            total_size = 0
            for k, v in value.items():
                r, size = Value.__to_yaml_truncated(v, limit - total_size)
                res[k] = r
                if limit - total_size < size:
                    res[TRUNCATED_TOKEN] = TRUNCATED_TOKEN
                    return res, limit
                total_size += size
            return res, total_size
        elif isinstance(value, tuple):
            return tuple(Value.__to_yaml_truncated_list(value, limit))
        else:
            return Value.__to_yaml_truncated_str(Value.to_str(value), limit)

    @staticmethod
    def to_yaml_truncated(value, limit: int = 5000):
        return Value.__to_yaml_truncated(value, limit)[0]

    @staticmethod
    def __to_yaml(value):
        if Value.is_primitive(value):
            return value
        elif isinstance(value, list):
            return [Value.to_yaml_truncated(v) for v in value]
        elif isinstance(value, dict):
            return {k: Value.to_yaml_truncated(v) for k, v in value.items()}
        elif isinstance(value, tuple):
            return tuple(Value.to_yaml_truncated(v) for v in value)
        else:
            return Value.to_str(value)

    @staticmethod
    def to_str(value):
        str_value = str(value)
        if str_value == Value.default_repr(value):
            if value.__class__.__module__ == '__main__':
                return value.__class__.__name__
            else:
                return f"{value.__class__.__module__}.{value.__class__.__name__}"
        else:
            return str_value

    @staticmethod
    def default_repr(value):
        return '<%s.%s object at %s>' % (
            value.__class__.__module__,
            value.__class__.__name__,
            hex(id(value))
        )


def get_config_names(decorator_name: str, function_name: str,
                     config_names: Union[str, 'ConfigItem', List['ConfigItem'], List[str]], *,
                     is_check_string_names: bool = True):
    from labml.internal.configs.config_item import ConfigItem
    if config_names is None:
        warnings.warn(f"Decorate your function with @{decorator_name}(Config.{function_name})", FutureWarning, 4)
        return function_name
    elif isinstance(config_names, str):
        if is_check_string_names:
            warnings.warn(f"Use Config.{config_names} instead of '{config_names}'.", FutureWarning, 4)
        return config_names
    elif isinstance(config_names, ConfigItem):
        return config_names.key
    else:
        assert isinstance(config_names, list) and len(config_names) > 0
        keys = []
        for c in config_names:
            if isinstance(c, str):
                warnings.warn(f"Use Config.{c} instead of '{c}'", FutureWarning, 4)
                keys.append(c)
            elif isinstance(c, ConfigItem):
                keys.append(c.key)
            else:
                raise ValueError('Cannot determine config item', c)
        return keys



from typing import Tuple, List, Any

from torch import nn
from torch.utils.data import DataLoader

from labml import experiment, tracker, monit
from labml.configs import option, calculate
from labml_helpers.metrics.collector import Collector
from labml_helpers.train_valid import SimpleTrainValidConfigs, BatchIndex
from labml_samples.pytorch.stocks.dataset import MinutelyData, MinutelyDataset
from labml_samples.pytorch.stocks.model import CnnModel


class Configs(SimpleTrainValidConfigs):
    epochs = 32
    dropout: float
    validation_dates: int = 100

    loss_func = nn.MSELoss()

    dataset: MinutelyData
    train_dataset: MinutelyDataset
    valid_dataset: MinutelyDataset
    model: CnnModel

    conv_sizes: List[Tuple[int, int]]
    activation: nn.Module

    train_batch_size: int = 32
    valid_batch_size: int = 64

    output_collector = Collector('output')

    def initialize(self):
        self.train_dataset = self.dataset.train_dataset()
        self.valid_dataset = self.dataset.valid_dataset()
        self.train_loader = DataLoader(self.train_dataset,
                                       batch_size=self.train_batch_size,
                                       shuffle=True)
        self.valid_loader = DataLoader(self.valid_dataset,
                                       batch_size=self.valid_batch_size,
                                       shuffle=False)
        self.state_modules = [self.output_collector]
        tracker.set_tensor('output.*')

    def step(self, batch: Any, batch_idx: BatchIndex):
        self.model.train(self.mode.is_train)
        data, target = batch['data'].to(self.device), batch['target'].to(self.device)
        target = (target - self.model.y_mean) / self.model.y_std

        if self.mode.is_train:
            tracker.add_global_step(len(data))

        output = self.model(data)
        loss = self.loss_func(output, target)
        tracker.add("loss.", loss)

        if self.mode.is_train:
            loss.backward()
            if batch_idx.is_last:
                tracker.add('model', self.model)
            self.optimizer.step()
            self.optimizer.zero_grad()

        if not self.mode.is_train:
            self.output_collector(output * self.model.y_std + self.model.y_mean)

        tracker.save()


@option(Configs.dataset)
def dataset(c: Configs):
    return MinutelyData(c.validation_dates)


@option(Configs.model)
def cnn_model(c: Configs):
    return CnnModel(price_mean=c.train_dataset.price_mean,
                    price_std=c.train_dataset.price_std,
                    volume_mean=c.train_dataset.volume_mean,
                    volume_std=c.train_dataset.volume_std,
                    y_mean=c.train_dataset.y_mean,
                    y_std=c.train_dataset.y_std,
                    activation=c.activation,
                    conv_sizes=c.conv_sizes,
                    dropout=c.dropout).to(c.device)


calculate(Configs.activation, 'relu', [], lambda: nn.ReLU())
calculate(Configs.activation, 'sigmoid', [], lambda: nn.Sigmoid())


def main():
    experiment.create()
    conf = Configs()
    conf.activation = 'relu'
    conf.dropout = 0.1
    experiment.configs(conf,
                       {'conv_sizes': [(128, 2), (256, 4)],
                        'optimizer.learning_rate': 1e-4,
                        'optimizer.optimizer': 'Adam'})

    with experiment.start():
        with monit.section('Initialize'):
            conf.initialize()
        with tracker.namespace('valid'):
            conf.valid_dataset.save_artifacts()
        conf.run()


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import Optional, Set, Dict, List, Union, TYPE_CHECKING, overload

import numpy as np

from labml.configs import BaseConfigs
from labml.internal.experiment import \
    create_experiment as _create_experiment, \
    experiment_singleton as _experiment_singleton, \
    ModelSaver
from labml.internal.experiment.experiment_run import \
    get_configs as _get_configs

if TYPE_CHECKING:
    import torch


def generate_uuid() -> str:
    from uuid import uuid1
    return uuid1().hex


def save_checkpoint():
    r"""
    Saves model checkpoints
    """
    _experiment_singleton().save_checkpoint()


def get_uuid():
    r"""
    Returns the UUID of the current experiment run
    """

    return _experiment_singleton().run.uuid


def create(*,
           uuid: Optional[str] = None,
           name: Optional[str] = None,
           python_file: Optional[str] = None,
           comment: Optional[str] = None,
           writers: Set[str] = None,
           ignore_callers: Set[str] = None,
           tags: Optional[Set[str]] = None):
    r"""
    Create an experiment

    Keyword Arguments:
        name (str, optional): name of the experiment
        python_file (str, optional): path of the Python file that
            created the experiment
        comment (str, optional): a short description of the experiment
        writers (Set[str], optional): list of writers to write stat to.
            Defaults to ``{'tensorboard', 'sqlite', 'web_api'}``.
        ignore_callers: (Set[str], optional): list of files to ignore when
            automatically determining ``python_file``
        tags (Set[str], optional): Set of tags for experiment
    """

    if writers is None:
        writers = {'screen', 'sqlite', 'tensorboard', 'web_api'}

    if ignore_callers is None:
        ignore_callers = set()

    if uuid is None:
        uuid = generate_uuid()

    _create_experiment(uuid=uuid,
                       name=name,
                       python_file=python_file,
                       comment=comment,
                       writers=writers,
                       ignore_callers=ignore_callers,
                       tags=tags,
                       is_evaluate=False)


def evaluate():
    r"""
    This should be used for evaluation of a saved experiment.
    This will not record anything.
    """

    _create_experiment(uuid=generate_uuid(),
                       name=None,
                       python_file=None,
                       comment=None,
                       writers=set(),
                       ignore_callers=set(),
                       tags=None,
                       is_evaluate=True)


def distributed(rank: int, world_size: int):
    _experiment_singleton().distributed(rank, world_size)


def add_model_savers(savers: Dict[str, ModelSaver]):
    _experiment_singleton().checkpoint_saver.add_savers(savers)


@overload
def add_pytorch_models(**kwargs: 'torch.nn.Module'):
    ...


@overload
def add_pytorch_models(models: Dict[str, 'torch.nn.Module']):
    ...


def add_pytorch_models(*args, **kwargs):
    """
    Set variables for saving and loading

    Arguments:
        models (Dict[str, torch.nn.Module]): a dictionary of torch modules
            used in the experiment.
            These will be saved with :func:`labml.experiment.save_checkpoint`
            and loaded with :func:`labml.experiment.load`.
    """

    from labml.internal.experiment.pytorch import add_models as _add_pytorch_models

    if len(args) > 0:
        assert len(args) == 1
        assert isinstance(args[0], dict)
        _add_pytorch_models(args[0])
    else:
        _add_pytorch_models(kwargs)


def add_sklearn_models(models: Dict[str, any]):
    """
    .. warning::
        This is still experimental.

    Set variables for saving and loading

    Arguments:
        models (Dict[str, any]): a dictionary of SKLearn models
            These will be saved with :func:`labml.experiment.save_checkpoint`
            and loaded with :func:`labml.experiment.load`.
    """
    from labml.internal.experiment.sklearn import add_models as _add_sklearn_models
    _add_sklearn_models(models)


@overload
def configs(conf_dict: Dict[str, any]):
    ...


@overload
def configs(conf_dict: Dict[str, any], conf_override: Dict[str, any]):
    ...


@overload
def configs(conf: BaseConfigs):
    ...


@overload
def configs(conf: BaseConfigs, conf_override: Dict[str, any]):
    ...


def configs(*args):
    r"""
    Calculate configurations

    This has multiple overloads

    .. function:: configs(conf_dict: Dict[str, any])
        :noindex:

    .. function:: configs(conf_dict: Dict[str, any], conf_override: Dict[str, any])
        :noindex:

    .. function:: configs(conf: BaseConfigs)
        :noindex:

    .. function:: configs(conf: BaseConfigs, run_order: List[Union[List[str], str]])
        :noindex:

    .. function:: configs(conf: BaseConfigs, *run_order: str)
        :noindex:

    .. function:: configs(conf: BaseConfigs, conf_override: Dict[str, any])
        :noindex:

    .. function:: configs(conf: BaseConfigs, conf_override: Dict[str, any], run_order: List[Union[List[str], str]])
        :noindex:

    .. function:: configs(conf: BaseConfigs, conf_override: Dict[str, any], *run_order: str)
        :noindex:

    Arguments:
        conf (BaseConfigs, optional): configurations object
        conf_dict (Dict[str, any], optional): a dictionary of configs
        conf_override (Dict[str, any], optional): a dictionary of
            configs to be overridden
    """
    conf_override: Optional[Dict[str, any]] = None
    conf = args[0]
    idx = 1

    if idx < len(args) and isinstance(args[idx], dict):
        conf_override = args[idx]
        idx += 1

    if len(args) != idx:
        raise RuntimeError("Invalid call to calculate configs")

    if isinstance(conf, BaseConfigs):
        _experiment_singleton().calc_configs(conf, conf_override)
    elif isinstance(conf, dict):
        _experiment_singleton().calc_configs(conf, conf_override)
    else:
        raise RuntimeError("Invalid call to calculate configs")


_load_run_uuid: Optional[str] = None
_load_checkpoint: Optional[int] = None


def start():
    r"""
    Starts the experiment.
    Run it using ``with`` statement and it will monitor and report, experiment completion
    and exceptions.
    """
    global _load_run_uuid
    global _load_checkpoint

    return _experiment_singleton().start(run_uuid=_load_run_uuid, checkpoint=_load_checkpoint)


def load_configs(run_uuid: str, *, is_only_hyperparam: bool = True):
    r"""
    Load configs of a previous run

    Arguments:
        run_uuid (str): if provided the experiment will start from
            a saved state in the run with UUID ``run_uuid``

    Keyword Arguments:
        is_only_hyperparam (bool, optional): if True all only the hyper parameters
            are returned
    """

    conf = _get_configs(run_uuid)
    if not conf:
        return {}
    values = {}
    for k, c in conf.items():
        is_hyperparam = c.get('is_hyperparam', None)
        is_explicit = c.get('is_explicitly_specified', False)

        if not is_only_hyperparam:
            values[k] = c['value']
        elif is_hyperparam is None and is_explicit:
            values[k] = c['value']
        elif is_hyperparam:
            values[k] = c['value']

    return values


def load(run_uuid: str, checkpoint: Optional[int] = None):
    r"""
    Loads a the run from a previous checkpoint.
    You need to separately call ``experiment.start`` to start the experiment.

    Arguments:
        run_uuid (str): experiment will start from
            a saved state in the run with UUID ``run_uuid``
        checkpoint (str, optional): if provided the experiment will start from
            given checkpoint. Otherwise it will start from the last checkpoint.
    """
    global _load_run_uuid
    global _load_checkpoint

    _load_run_uuid = run_uuid
    _load_checkpoint = checkpoint


def load_models(models: List[str], run_uuid: str, checkpoint: Optional[int] = None):
    r"""
    Loads and starts the run from a previous checkpoint.

    Arguments:
        models (List[str]): List of names of models to be loaded
        run_uuid (str): experiment will start from
            a saved state in the run with UUID ``run_uuid``
        checkpoint (str, optional): if provided the experiment will start from
            given checkpoint. Otherwise it will start from the last checkpoint.
    """

    _experiment_singleton().load_models(models=models, run_uuid=run_uuid, checkpoint=checkpoint)


def save_numpy(name: str, array: np.ndarray):
    r"""
    Saves a single numpy array. This is used to save processed data.
    """

    numpy_path = Path(_experiment_singleton().run.numpy_path)

    if not numpy_path.exists():
        numpy_path.mkdir(parents=True)
    file_name = name + ".npy"
    np.save(str(numpy_path / file_name), array)


def record(*,
           name: Optional[str] = None,
           comment: Optional[str] = None,
           writers: Set[str] = None,
           tags: Optional[Set[str]] = None,
           exp_conf: Dict[str, any] = None,
           lab_conf: Dict[str, any] = None,
           token: str = None):
    r"""
    This is combines :func:`create`, :func:`configs` and :func:`start`.

    Keyword Arguments:
        name (str, optional): name of the experiment
        comment (str, optional): a short description of the experiment
        writers (Set[str], optional): list of writers to write stat to.
            Defaults to ``{'tensorboard', 'sqlite', 'web_api'}``.
        tags (Set[str], optional): Set of tags for experiment
        exp_conf (Dict[str, any], optional): a dictionary of experiment configurations
        lab_conf (Dict[str, any], optional): a dictionary of configurations for LabML.
         Use this if you want to change default configurations such as ``web_api``, and
         ``data_path``.
        token (str, optional): a shortcut to provide LabML mobile app token (or url - ``web_api``)
         instead of including it in ``lab_conf``. You can set this with :func:`labml.lab.configure`,
         `or with a configuration file for the entire project <../guide/installation_setup.html>`_.
    """

    if token is not None:
        if lab_conf is None:
            lab_conf = {}
        lab_conf['web_api'] = token

    if lab_conf is not None:
        from labml.internal.lab import lab_singleton as _internal
        _internal().set_configurations(lab_conf)

    create(name=name,
           python_file=None,
           comment=comment,
           writers=writers,
           ignore_callers=None,
           tags=tags)

    if exp_conf is not None:
        configs(exp_conf)

    return start()



import sqlite3
from typing import Optional

import numpy as np

from .analytics import Analytics, BASIS_POINTS


def _filter_steps(start_step: Optional[int], end_step: Optional[int]):
    sql = ''
    if start_step is not None:
        sql += f' AND step >= {start_step}'
    if end_step is not None:
        sql += f' AND step < {end_step}'

    return sql


class SQLiteAnalytics(Analytics):
    def __init__(self, sqlite_path):
        self.conn = sqlite3.connect(str(sqlite_path))

    def get_key(self, name):
        return name

    def scalar(self, name: str, start_step: Optional[int], end_step: Optional[int]):
        key = self.get_key(name)
        sql = f'SELECT step, value from scalars WHERE indicator = "{key}"'
        sql += _filter_steps(start_step, end_step)
        cur = self.conn.execute(sql)
        return [c for c in cur]

    def summarize(self, events):
        step = np.mean([e[0] for e in events])
        values = np.sort([e[1] for e in events])
        basis_points = np.percentile(values, BASIS_POINTS)

        return np.concatenate(([step], basis_points))

    def tensor(self, name: str, start_step: Optional[int], end_step: Optional[int]):
        key = self.get_key(name)
        sql = f'SELECT step, filename from tensors WHERE indicator = "{key}"'
        sql += _filter_steps(start_step, end_step)
        cur = self.conn.execute(sql)
        return [c for c in cur]
